#***************************************************************************
#*   Copyright (c) 2001, 2002 Juergen Riegel <juergen.riegel@web.de>       *
#*                                                                         *
#*   This file is part of the FreeCAD CAx development system.              *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   FreeCAD is distributed in the hope that it will be useful,            *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Lesser General Public License for more details.                   *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with FreeCAD; if not, write to the Free Software        *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************/

# FreeCAD init script of the part module

#FreeCAD.addImportType("CAD formats (*.igs *.iges *.step *.stp *.brep *.brp)","Part")
#FreeCAD.addExportType("CAD formats (*.igs *.iges *.step *.stp *.brep *.brp)","Part")
FreeCAD.addImportType("BREP format (*.brep *.BREP *.brp *.BRP)","Part")
FreeCAD.addExportType("BREP format (*.brep *.brp)","Part")
FreeCAD.addImportType("IGES format (*.iges *.IGES *.igs *.IGS)","Part")
FreeCAD.addExportType("IGES format (*.iges *.igs)","Part")
FreeCAD.addImportType("STEP with colors (*.step *.STEP *.stp *.STP)","Import")
FreeCAD.addExportType("STEP with colors (*.step *.stp)","Import")

FreeCAD.__unit_test__ += [ "TestPartApp" ]



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Contains the unit test class for addonmanager_installer.py non-GUI functionality."""

import unittest
from unittest.mock import Mock
import os
import shutil
import tempfile
from zipfile import ZipFile
import sys

sys.path.append("../../")  # So the IDE can find the imports below

import FreeCAD
from addonmanager_installer import InstallationMethod, AddonInstaller, MacroInstaller
from addonmanager_git import GitManager, initialize_git
from addonmanager_metadata import MetadataReader
from Addon import Addon
from AddonManagerTest.app.mocks import MockAddon, MockMacro


class TestAddonInstaller(unittest.TestCase):
    """Test class for addonmanager_installer.py non-GUI functionality"""

    MODULE = "test_installer"  # file name without extension

    def setUp(self):
        """Initialize data needed for all tests"""
        # self.start_time = time.perf_counter()
        self.test_data_dir = os.path.join(
            FreeCAD.getHomePath(), "Mod", "AddonManager", "AddonManagerTest", "data"
        )
        self.real_addon = Addon(
            "TestAddon",
            "https://github.com/FreeCAD/FreeCAD-addons",
            Addon.Status.NOT_INSTALLED,
            "master",
        )
        self.mock_addon = MockAddon()

    def tearDown(self):
        """Finalize the test."""
        # end_time = time.perf_counter()
        # print(f"Test '{self.id()}' ran in {end_time-self.start_time:.4f} seconds")

    def test_validate_object(self):
        """An object is valid if it has a name, url, and branch attribute."""

        AddonInstaller._validate_object(self.real_addon)  # Won't raise
        AddonInstaller._validate_object(self.mock_addon)  # Won't raise

        class NoName:
            def __init__(self):
                self.url = "https://github.com/FreeCAD/FreeCAD-addons"
                self.branch = "master"

        no_name = NoName()
        with self.assertRaises(RuntimeError):
            AddonInstaller._validate_object(no_name)

        class NoUrl:
            def __init__(self):
                self.name = "TestAddon"
                self.branch = "master"

        no_url = NoUrl()
        with self.assertRaises(RuntimeError):
            AddonInstaller._validate_object(no_url)

        class NoBranch:
            def __init__(self):
                self.name = "TestAddon"
                self.url = "https://github.com/FreeCAD/FreeCAD-addons"

        no_branch = NoBranch()
        with self.assertRaises(RuntimeError):
            AddonInstaller._validate_object(no_branch)

    def test_update_metadata(self):
        """If a metadata file exists in the installation location, it should be
        loaded."""
        addon = Mock()
        addon.name = "MockAddon"
        installer = AddonInstaller(addon, [])
        installer._update_metadata()  # Does nothing, but should not crash

        installer = AddonInstaller(self.real_addon, [])
        with tempfile.TemporaryDirectory() as temp_dir:
            installer.installation_path = temp_dir
            installer._update_metadata()
            addon_dir = os.path.join(temp_dir, self.real_addon.name)
            os.mkdir(addon_dir)
            shutil.copy(
                os.path.join(self.test_data_dir, "good_package.xml"),
                os.path.join(addon_dir, "package.xml"),
            )
            good_metadata = MetadataReader.from_file(
                os.path.join(addon_dir, "package.xml")
            )
            installer._update_metadata()
            self.assertEqual(self.real_addon.installed_version, good_metadata.version)

    def test_finalize_zip_installation_non_github(self):
        """Ensure that zip files are correctly extracted."""
        with tempfile.TemporaryDirectory() as temp_dir:
            test_simple_repo = os.path.join(self.test_data_dir, "test_simple_repo.zip")
            non_gh_mock = MockAddon()
            non_gh_mock.url = test_simple_repo
            non_gh_mock.name = "NonGitHubMock"
            installer = AddonInstaller(non_gh_mock, [])
            installer.installation_path = temp_dir
            installer._finalize_zip_installation(test_simple_repo)
            expected_location = os.path.join(temp_dir, non_gh_mock.name, "README")
            self.assertTrue(
                os.path.isfile(expected_location), "Non-GitHub zip extraction failed"
            )

    def test_finalize_zip_installation_github(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            test_github_style_repo = os.path.join(
                self.test_data_dir, "test_github_style_repo.zip"
            )
            self.mock_addon.url = test_github_style_repo
            self.mock_addon.branch = "master"
            installer = AddonInstaller(self.mock_addon, [])
            installer.installation_path = temp_dir
            installer._finalize_zip_installation(test_github_style_repo)
            expected_location = os.path.join(temp_dir, self.mock_addon.name, "README")
            self.assertTrue(
                os.path.isfile(expected_location), "GitHub zip extraction failed"
            )

    def test_code_in_branch_subdirectory_true(self):
        """When there is a subdirectory with the branch name in it, find it"""
        installer = AddonInstaller(self.mock_addon, [])
        with tempfile.TemporaryDirectory() as temp_dir:
            os.mkdir(
                os.path.join(
                    temp_dir, f"{self.mock_addon.name}-{self.mock_addon.branch}"
                )
            )
            result = installer._code_in_branch_subdirectory(temp_dir)
            self.assertTrue(result, "Failed to find ZIP subdirectory")

    def test_code_in_branch_subdirectory_false(self):
        """When there is not a subdirectory with the branch name in it, don't find
        one"""
        installer = AddonInstaller(self.mock_addon, [])
        with tempfile.TemporaryDirectory() as temp_dir:
            result = installer._code_in_branch_subdirectory(temp_dir)
            self.assertFalse(result, "Found ZIP subdirectory when there was none")

    def test_code_in_branch_subdirectory_more_than_one(self):
        """When there are multiple subdirectories, never find a branch subdirectory"""
        installer = AddonInstaller(self.mock_addon, [])
        with tempfile.TemporaryDirectory() as temp_dir:
            os.mkdir(
                os.path.join(
                    temp_dir, f"{self.mock_addon.name}-{self.mock_addon.branch}"
                )
            )
            os.mkdir(os.path.join(temp_dir, "AnotherSubdir"))
            result = installer._code_in_branch_subdirectory(temp_dir)
            self.assertFalse(
                result, "Found ZIP subdirectory when there were multiple subdirs"
            )

    def test_move_code_out_of_subdirectory(self):
        """All files are moved out and the subdirectory is deleted"""
        installer = AddonInstaller(self.mock_addon, [])
        with tempfile.TemporaryDirectory() as temp_dir:
            subdir = os.path.join(
                temp_dir, f"{self.mock_addon.name}-{self.mock_addon.branch}"
            )
            os.mkdir(subdir)
            with open(os.path.join(subdir, "README.txt"), "w", encoding="utf-8") as f:
                f.write("# Test file for unit testing")
            with open(
                os.path.join(subdir, "AnotherFile.txt"), "w", encoding="utf-8"
            ) as f:
                f.write("# Test file for unit testing")
            installer._move_code_out_of_subdirectory(temp_dir)
            self.assertTrue(os.path.isfile(os.path.join(temp_dir, "README.txt")))
            self.assertTrue(os.path.isfile(os.path.join(temp_dir, "AnotherFile.txt")))
            self.assertFalse(os.path.isdir(subdir))

    def test_install_by_git(self):
        """Test using git to install. Depends on there being a local git
        installation: the test is skipped if there is no local git."""
        git_manager = initialize_git()
        if not git_manager:
            self.skipTest("git not found, skipping git installer tests")
            return

        # Our test git repo has to be in a zipfile, otherwise it cannot itself be
        # stored in git, since it has a .git subdirectory.
        with tempfile.TemporaryDirectory() as temp_dir:
            git_repo_zip = os.path.join(self.test_data_dir, "test_repo.zip")
            with ZipFile(git_repo_zip, "r") as zip_repo:
                zip_repo.extractall(temp_dir)

            mock_addon = MockAddon()
            mock_addon.url = os.path.join(temp_dir, "test_repo")
            mock_addon.branch = "main"
            installer = AddonInstaller(mock_addon, [])
            installer.installation_path = os.path.join(temp_dir, "installed_addon")
            installer._install_by_git()

            self.assertTrue(os.path.exists(installer.installation_path))
            addon_name_dir = os.path.join(installer.installation_path, mock_addon.name)
            self.assertTrue(os.path.exists(addon_name_dir))
            readme = os.path.join(addon_name_dir, "README.md")
            self.assertTrue(os.path.exists(readme))

    def test_install_by_copy(self):
        """Test using a simple filesystem copy to install an addon."""
        with tempfile.TemporaryDirectory() as temp_dir:
            git_repo_zip = os.path.join(self.test_data_dir, "test_repo.zip")
            with ZipFile(git_repo_zip, "r") as zip_repo:
                zip_repo.extractall(temp_dir)

            mock_addon = MockAddon()
            mock_addon.url = os.path.join(temp_dir, "test_repo")
            mock_addon.branch = "main"
            installer = AddonInstaller(mock_addon, [])
            installer.addon_to_install = mock_addon
            installer.installation_path = os.path.join(temp_dir, "installed_addon")
            installer._install_by_copy()

            self.assertTrue(os.path.exists(installer.installation_path))
            addon_name_dir = os.path.join(installer.installation_path, mock_addon.name)
            self.assertTrue(os.path.exists(addon_name_dir))
            readme = os.path.join(addon_name_dir, "README.md")
            self.assertTrue(os.path.exists(readme))

    def test_determine_install_method_local_path(self):
        """Test which install methods are accepted for a local path"""

        with tempfile.TemporaryDirectory() as temp_dir:
            installer = AddonInstaller(self.mock_addon, [])
            method = installer._determine_install_method(
                temp_dir, InstallationMethod.COPY
            )
            self.assertEqual(method, InstallationMethod.COPY)
            git_manager = initialize_git()
            if git_manager:
                method = installer._determine_install_method(
                    temp_dir, InstallationMethod.GIT
                )
                self.assertEqual(method, InstallationMethod.GIT)
            method = installer._determine_install_method(
                temp_dir, InstallationMethod.ZIP
            )
            self.assertIsNone(method)
            method = installer._determine_install_method(
                temp_dir, InstallationMethod.ANY
            )
            self.assertEqual(method, InstallationMethod.COPY)

    def test_determine_install_method_file_url(self):
        """Test which install methods are accepted for a file:// url"""

        with tempfile.TemporaryDirectory() as temp_dir:
            installer = AddonInstaller(self.mock_addon, [])
            temp_dir = "file://" + temp_dir.replace(os.path.sep, "/")
            method = installer._determine_install_method(
                temp_dir, InstallationMethod.COPY
            )
            self.assertEqual(method, InstallationMethod.COPY)
            git_manager = initialize_git()
            if git_manager:
                method = installer._determine_install_method(
                    temp_dir, InstallationMethod.GIT
                )
                self.assertEqual(method, InstallationMethod.GIT)
            method = installer._determine_install_method(
                temp_dir, InstallationMethod.ZIP
            )
            self.assertIsNone(method)
            method = installer._determine_install_method(
                temp_dir, InstallationMethod.ANY
            )
            self.assertEqual(method, InstallationMethod.COPY)

    def test_determine_install_method_local_zip(self):
        """Test which install methods are accepted for a local path to a zipfile"""

        with tempfile.TemporaryDirectory() as temp_dir:
            installer = AddonInstaller(self.mock_addon, [])
            temp_file = os.path.join(temp_dir, "dummy.zip")
            method = installer._determine_install_method(
                temp_file, InstallationMethod.COPY
            )
            self.assertEqual(method, InstallationMethod.ZIP)
            method = installer._determine_install_method(
                temp_file, InstallationMethod.GIT
            )
            self.assertIsNone(method)
            method = installer._determine_install_method(
                temp_file, InstallationMethod.ZIP
            )
            self.assertEqual(method, InstallationMethod.ZIP)
            method = installer._determine_install_method(
                temp_file, InstallationMethod.ANY
            )
            self.assertEqual(method, InstallationMethod.ZIP)

    def test_determine_install_method_remote_zip(self):
        """Test which install methods are accepted for a remote path to a zipfile"""

        installer = AddonInstaller(self.mock_addon, [])

        temp_file = "https://freecad.org/dummy.zip"  # Doesn't have to actually exist!

        method = installer._determine_install_method(temp_file, InstallationMethod.COPY)
        self.assertIsNone(method)
        method = installer._determine_install_method(temp_file, InstallationMethod.GIT)
        self.assertIsNone(method)
        method = installer._determine_install_method(temp_file, InstallationMethod.ZIP)
        self.assertEqual(method, InstallationMethod.ZIP)
        method = installer._determine_install_method(temp_file, InstallationMethod.ANY)
        self.assertEqual(method, InstallationMethod.ZIP)

    def test_determine_install_method_https_known_sites_copy(self):
        """Test which install methods are accepted for an https GitHub URL"""

        installer = AddonInstaller(self.mock_addon, [])
        installer.git_manager = True

        for site in ["github.org", "gitlab.org", "framagit.org", "salsa.debian.org"]:
            with self.subTest(site=site):
                temp_file = (
                    f"https://{site}/dummy/dummy"  # Doesn't have to actually exist!
                )
                method = installer._determine_install_method(
                    temp_file, InstallationMethod.COPY
                )
                self.assertIsNone(method, f"Allowed copying from {site} URL")

    def test_determine_install_method_https_known_sites_git(self):
        """Test which install methods are accepted for an https GitHub URL"""

        installer = AddonInstaller(self.mock_addon, [])
        installer.git_manager = True

        for site in ["github.org", "gitlab.org", "framagit.org", "salsa.debian.org"]:
            with self.subTest(site=site):
                temp_file = (
                    f"https://{site}/dummy/dummy"  # Doesn't have to actually exist!
                )
                method = installer._determine_install_method(
                    temp_file, InstallationMethod.GIT
                )
                self.assertEqual(
                    method,
                    InstallationMethod.GIT,
                    f"Failed to allow git access to {site} URL",
                )

    def test_determine_install_method_https_known_sites_zip(self):
        """Test which install methods are accepted for an https GitHub URL"""

        installer = AddonInstaller(self.mock_addon, [])
        installer.git_manager = True

        for site in ["github.org", "gitlab.org", "framagit.org", "salsa.debian.org"]:
            with self.subTest(site=site):
                temp_file = (
                    f"https://{site}/dummy/dummy"  # Doesn't have to actually exist!
                )
                method = installer._determine_install_method(
                    temp_file, InstallationMethod.ZIP
                )
                self.assertEqual(
                    method,
                    InstallationMethod.ZIP,
                    f"Failed to allow zip access to {site} URL",
                )

    def test_determine_install_method_https_known_sites_any_gm(self):
        """Test which install methods are accepted for an https GitHub URL"""

        installer = AddonInstaller(self.mock_addon, [])
        installer.git_manager = True

        for site in ["github.org", "gitlab.org", "framagit.org", "salsa.debian.org"]:
            with self.subTest(site=site):
                temp_file = (
                    f"https://{site}/dummy/dummy"  # Doesn't have to actually exist!
                )
                method = installer._determine_install_method(
                    temp_file, InstallationMethod.ANY
                )
                self.assertEqual(
                    method,
                    InstallationMethod.GIT,
                    f"Failed to allow git access to {site} URL",
                )

    def test_determine_install_method_https_known_sites_any_no_gm(self):
        """Test which install methods are accepted for an https GitHub URL"""

        installer = AddonInstaller(self.mock_addon, [])
        installer.git_manager = None

        for site in ["github.org", "gitlab.org", "framagit.org", "salsa.debian.org"]:
            with self.subTest(site=site):
                temp_file = (
                    f"https://{site}/dummy/dummy"  # Doesn't have to actually exist!
                )
                method = installer._determine_install_method(
                    temp_file, InstallationMethod.ANY
                )
                self.assertEqual(
                    method,
                    InstallationMethod.ZIP,
                    f"Failed to allow zip access to {site} URL",
                )

    def test_fcmacro_copying(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            mock_addon = MockAddon()
            mock_addon.url = os.path.join(
                self.test_data_dir, "test_addon_with_fcmacro.zip"
            )
            installer = AddonInstaller(mock_addon, [])
            installer.installation_path = temp_dir
            installer.macro_installation_path = os.path.join(temp_dir, "Macros")
            installer.run()
            self.assertTrue(
                os.path.exists(os.path.join(temp_dir, "Macros", "TestMacro.FCMacro")),
                "FCMacro file was not copied to macro installation location",
            )


class TestMacroInstaller(unittest.TestCase):
    MODULE = "test_installer"  # file name without extension

    def setUp(self):
        """Set up the mock objects"""

        self.mock = MockAddon()
        self.mock.macro = MockMacro()

    def test_installation(self):
        """Test the wrapper around the macro installer"""

        # Note that this doesn't test the underlying Macro object's install function,
        # it only tests whether that function is called appropriately by the
        # MacroInstaller wrapper.
        with tempfile.TemporaryDirectory() as temp_dir:
            installer = MacroInstaller(self.mock)
            installer.installation_path = temp_dir
            installation_succeeded = installer.run()
            self.assertTrue(installation_succeeded)
            self.assertTrue(
                os.path.exists(os.path.join(temp_dir, self.mock.macro.filename))
            )



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2023 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Tests for the Addon Manager's FreeCAD interface classes."""

import json
import os
import sys
import tempfile
import unittest
from unittest.mock import patch, MagicMock

# pylint: disable=protected-access,import-outside-toplevel


class TestConsole(unittest.TestCase):
    """Tests for the Console"""

    def setUp(self) -> None:
        self.saved_freecad = None
        if "FreeCAD" in sys.modules:
            self.saved_freecad = sys.modules["FreeCAD"]
            sys.modules.pop("FreeCAD")
        if "addonmanager_freecad_interface" in sys.modules:
            sys.modules.pop("addonmanager_freecad_interface")
        sys.path.append("../../")

    def tearDown(self) -> None:
        if "FreeCAD" in sys.modules:
            sys.modules.pop("FreeCAD")
        if self.saved_freecad is not None:
            sys.modules["FreeCAD"] = self.saved_freecad

    def test_log_with_freecad(self):
        """Ensure that if FreeCAD exists, the appropriate function is called"""
        sys.modules["FreeCAD"] = unittest.mock.MagicMock()
        import addonmanager_freecad_interface as fc

        fc.Console.PrintLog("Test output")
        self.assertTrue(isinstance(fc.Console, unittest.mock.MagicMock))
        self.assertTrue(fc.Console.PrintLog.called)

    def test_log_no_freecad(self):
        """Test that if the FreeCAD import fails, the logger is set up correctly, and
        implements PrintLog"""
        sys.modules["FreeCAD"] = None
        with patch(
            "addonmanager_freecad_interface.logging", new=MagicMock()
        ) as mock_logging:
            import addonmanager_freecad_interface as fc

            fc.Console.PrintLog("Test output")
            self.assertTrue(isinstance(fc.Console, fc.ConsoleReplacement))
            self.assertTrue(mock_logging.log.called)

    def test_message_no_freecad(self):
        """Test that if the FreeCAD import fails the logger implements PrintMessage"""
        sys.modules["FreeCAD"] = None
        with patch(
            "addonmanager_freecad_interface.logging", new=MagicMock()
        ) as mock_logging:
            import addonmanager_freecad_interface as fc

            fc.Console.PrintMessage("Test output")
            self.assertTrue(mock_logging.info.called)

    def test_warning_no_freecad(self):
        """Test that if the FreeCAD import fails the logger implements PrintWarning"""
        sys.modules["FreeCAD"] = None
        with patch(
            "addonmanager_freecad_interface.logging", new=MagicMock()
        ) as mock_logging:
            import addonmanager_freecad_interface as fc

            fc.Console.PrintWarning("Test output")
            self.assertTrue(mock_logging.warning.called)

    def test_error_no_freecad(self):
        """Test that if the FreeCAD import fails the logger implements PrintError"""
        sys.modules["FreeCAD"] = None
        with patch(
            "addonmanager_freecad_interface.logging", new=MagicMock()
        ) as mock_logging:
            import addonmanager_freecad_interface as fc

            fc.Console.PrintError("Test output")
            self.assertTrue(mock_logging.error.called)


class TestParameters(unittest.TestCase):
    """Tests for the Parameters"""

    def setUp(self) -> None:
        self.saved_freecad = None
        if "FreeCAD" in sys.modules:
            self.saved_freecad = sys.modules["FreeCAD"]
            sys.modules.pop("FreeCAD")
        if "addonmanager_freecad_interface" in sys.modules:
            sys.modules.pop("addonmanager_freecad_interface")
        sys.path.append("../../")

    def tearDown(self) -> None:
        if "FreeCAD" in sys.modules:
            sys.modules.pop("FreeCAD")
        if self.saved_freecad is not None:
            sys.modules["FreeCAD"] = self.saved_freecad

    def test_param_get_with_freecad(self):
        """Ensure that if FreeCAD exists, the built-in FreeCAD function is called"""
        sys.modules["FreeCAD"] = unittest.mock.MagicMock()
        import addonmanager_freecad_interface as fc

        prefs = fc.ParamGet("some/fake/path")
        self.assertTrue(isinstance(prefs, unittest.mock.MagicMock))

    def test_param_get_no_freecad(self):
        """Test that if the FreeCAD import fails, param_get returns a ParametersReplacement"""
        sys.modules["FreeCAD"] = None
        import addonmanager_freecad_interface as fc

        prefs = fc.ParamGet("some/fake/path")
        self.assertTrue(isinstance(prefs, fc.ParametersReplacement))

    def test_replacement_getters_and_setters(self):
        """Test that ParameterReplacement's getters, setters, and deleters work"""
        sys.modules["FreeCAD"] = None
        import addonmanager_freecad_interface as fc

        prf = fc.ParamGet("some/fake/path")
        gs_types = [
            ("Bool", prf.GetBool, prf.SetBool, prf.RemBool, True, False),
            ("Int", prf.GetInt, prf.SetInt, prf.RemInt, 42, 0),
            ("Float", prf.GetFloat, prf.SetFloat, prf.RemFloat, 1.2, 3.4),
            ("String", prf.GetString, prf.SetString, prf.RemString, "test", "other"),
        ]
        for gs_type in gs_types:
            with self.subTest(msg=f"Testing {gs_type[0]}", gs_type=gs_type):
                getter = gs_type[1]
                setter = gs_type[2]
                deleter = gs_type[3]
                value_1 = gs_type[4]
                value_2 = gs_type[5]
                self.assertEqual(getter("test", value_1), value_1)
                self.assertEqual(getter("test", value_2), value_2)
                self.assertNotIn("test", prf.parameters)
                setter("test", value_1)
                self.assertIn("test", prf.parameters)
                self.assertEqual(getter("test", value_2), value_1)
                deleter("test")
                self.assertNotIn("test", prf.parameters)


class TestDataPaths(unittest.TestCase):
    """Tests for the data paths"""

    def setUp(self) -> None:
        self.saved_freecad = None
        if "FreeCAD" in sys.modules:
            self.saved_freecad = sys.modules["FreeCAD"]
            sys.modules.pop("FreeCAD")
        if "addonmanager_freecad_interface" in sys.modules:
            sys.modules.pop("addonmanager_freecad_interface")
        sys.path.append("../../")

    def tearDown(self) -> None:
        if "FreeCAD" in sys.modules:
            sys.modules.pop("FreeCAD")
        if self.saved_freecad is not None:
            sys.modules["FreeCAD"] = self.saved_freecad

    def test_init_with_freecad(self):
        """Ensure that if FreeCAD exists, the appropriate functions are called"""
        sys.modules["FreeCAD"] = unittest.mock.MagicMock()
        import addonmanager_freecad_interface as fc

        data_paths = fc.DataPaths()
        self.assertTrue(sys.modules["FreeCAD"].getUserAppDataDir.called)
        self.assertTrue(sys.modules["FreeCAD"].getUserMacroDir.called)
        self.assertTrue(sys.modules["FreeCAD"].getUserCachePath.called)
        self.assertIsNotNone(data_paths.mod_dir)
        self.assertIsNotNone(data_paths.cache_dir)
        self.assertIsNotNone(data_paths.macro_dir)

    def test_init_without_freecad(self):
        """Ensure that if FreeCAD does not exist, the appropriate functions are called"""
        sys.modules["FreeCAD"] = None
        import addonmanager_freecad_interface as fc

        data_paths = fc.DataPaths()
        self.assertIsNotNone(data_paths.mod_dir)
        self.assertIsNotNone(data_paths.cache_dir)
        self.assertIsNotNone(data_paths.macro_dir)
        self.assertNotEqual(data_paths.mod_dir, data_paths.cache_dir)
        self.assertNotEqual(data_paths.mod_dir, data_paths.macro_dir)
        self.assertNotEqual(data_paths.cache_dir, data_paths.macro_dir)


class TestPreferences(unittest.TestCase):
    """Tests for the preferences wrapper"""

    def setUp(self) -> None:
        sys.path.append("../../")
        import addonmanager_freecad_interface as fc

        self.fc = fc

    def tearDown(self) -> None:
        pass

    def test_load_preferences_defaults(self):
        """Preferences are loaded from a given file"""
        defaults = self.given_defaults()
        with tempfile.TemporaryDirectory() as temp_dir:
            json_file = os.path.join(temp_dir, "defaults.json")
            with open(json_file, "w", encoding="utf-8") as f:
                f.write(json.dumps(defaults))
            self.fc.Preferences._load_preferences_defaults(json_file)
            self.assertDictEqual(defaults, self.fc.Preferences.preferences_defaults)

    def test_in_memory_defaults(self):
        """Preferences are loaded from memory"""
        defaults = self.given_defaults()
        prefs = self.fc.Preferences(defaults)
        self.assertDictEqual(defaults, prefs.preferences_defaults)

    def test_get_good(self):
        """Get returns results when matching an existing preference"""
        defaults = self.given_defaults()
        prefs = self.fc.Preferences(defaults)
        self.assertEqual(prefs.get("TestBool"), defaults["TestBool"])
        self.assertEqual(prefs.get("TestInt"), defaults["TestInt"])
        self.assertEqual(prefs.get("TestFloat"), defaults["TestFloat"])
        self.assertEqual(prefs.get("TestString"), defaults["TestString"])

    def test_get_nonexistent(self):
        """Get raises an exception when asked for a non-existent preference"""
        defaults = self.given_defaults()
        prefs = self.fc.Preferences(defaults)
        with self.assertRaises(RuntimeError):
            prefs.get("No_such_thing")

    def test_get_bad_type(self):
        """Get raises an exception when getting an unsupported type"""
        defaults = self.given_defaults()
        defaults["TestArray"] = ["This", "Is", "Legal", "JSON"]
        prefs = self.fc.Preferences(defaults)
        with self.assertRaises(RuntimeError):
            prefs.get("TestArray")

    def test_set_good(self):
        """Set works when matching an existing preference"""
        defaults = self.given_defaults()
        prefs = self.fc.Preferences(defaults)
        prefs.set("TestBool", False)
        self.assertEqual(prefs.get("TestBool"), False)
        prefs.set("TestInt", 4321)
        self.assertEqual(prefs.get("TestInt"), 4321)
        prefs.set("TestFloat", 3.14159)
        self.assertEqual(prefs.get("TestFloat"), 3.14159)
        prefs.set("TestString", "Forty two")
        self.assertEqual(prefs.get("TestString"), "Forty two")

    def test_set_nonexistent(self):
        """Set raises an exception when asked for a non-existent preference"""
        defaults = self.given_defaults()
        prefs = self.fc.Preferences(defaults)
        with self.assertRaises(RuntimeError):
            prefs.get("No_such_thing")

    def test_set_bad_type(self):
        """Set raises an exception when setting an unsupported type"""
        defaults = self.given_defaults()
        defaults["TestArray"] = ["This", "Is", "Legal", "JSON"]
        prefs = self.fc.Preferences(defaults)
        with self.assertRaises(RuntimeError):
            prefs.get("TestArray")

    @staticmethod
    def given_defaults():
        """Get a dictionary of fake defaults for testing"""
        defaults = {
            "TestBool": True,
            "TestInt": 42,
            "TestFloat": 1.2,
            "TestString": "Test",
        }
        return defaults



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Contains a class for adding a single content item, as well as auxiliary classes for
its dependent dialog boxes. """

import os
from typing import Optional, Tuple, List

import FreeCAD
import FreeCADGui

from Addon import INTERNAL_WORKBENCHES

from PySide.QtWidgets import (
    QDialog,
    QLayout,
    QFileDialog,
    QTableWidgetItem,
    QSizePolicy,
)
from PySide.QtGui import QIcon
from PySide.QtCore import Qt

from addonmanager_devmode_validators import (
    VersionValidator,
    NameValidator,
    PythonIdentifierValidator,
)
from addonmanager_devmode_people_table import PeopleTable
from addonmanager_devmode_licenses_table import LicensesTable

# pylint: disable=too-few-public-methods

translate = FreeCAD.Qt.translate


class AddContent:
    """A dialog for adding a single content item to the package metadata."""

    def __init__(self, path_to_addon: str, toplevel_metadata: FreeCAD.Metadata):
        """path_to_addon is the full path to the toplevel directory of this Addon, and
        toplevel_metadata is to overall package.xml Metadata object for this Addon. This
        information is used to assist the use in filling out the dialog by providing
        sensible default values."""
        self.dialog = FreeCADGui.PySideUic.loadUi(
            os.path.join(os.path.dirname(__file__), "developer_mode_add_content.ui")
        )
        # These are in alphabetical order in English, but their actual label may be translated in
        # the GUI. Store their underlying type as user data.
        self.dialog.addonKindComboBox.setItemData(0, "macro")
        self.dialog.addonKindComboBox.setItemData(1, "preferencepack")
        self.dialog.addonKindComboBox.setItemData(2, "workbench")

        self.people_table = PeopleTable()
        self.licenses_table = LicensesTable()
        large_size_policy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        large_size_policy.setHorizontalStretch(2)
        small_size_policy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        small_size_policy.setHorizontalStretch(1)
        self.people_table.widget.setSizePolicy(large_size_policy)
        self.licenses_table.widget.setSizePolicy(small_size_policy)
        self.dialog.peopleAndLicenseshorizontalLayout.addWidget(
            self.people_table.widget
        )
        self.dialog.peopleAndLicenseshorizontalLayout.addWidget(
            self.licenses_table.widget
        )

        self.toplevel_metadata = toplevel_metadata
        self.metadata = None
        self.path_to_addon = path_to_addon.replace("/", os.path.sep)
        if self.path_to_addon[-1] != os.path.sep:
            self.path_to_addon += (
                os.path.sep
            )  # Make sure the path ends with a separator

        self.dialog.iconLabel.hide()  # Until we have an icon to display

        self.dialog.iconBrowseButton.clicked.connect(self._browse_for_icon_clicked)
        self.dialog.subdirectoryBrowseButton.clicked.connect(
            self._browse_for_subdirectory_clicked
        )
        self.dialog.tagsButton.clicked.connect(self._tags_clicked)
        self.dialog.dependenciesButton.clicked.connect(self._dependencies_clicked)
        self.dialog.freecadVersionsButton.clicked.connect(
            self._freecad_versions_clicked
        )

        self.dialog.versionLineEdit.setValidator(VersionValidator())
        self.dialog.prefPackNameLineEdit.setValidator(NameValidator())
        self.dialog.displayNameLineEdit.setValidator(NameValidator())
        self.dialog.workbenchClassnameLineEdit.setValidator(PythonIdentifierValidator())

    def exec(
        self,
        content_kind: str = "workbench",
        metadata: FreeCAD.Metadata = None,
        singleton: bool = True,
    ) -> Optional[Tuple[str, FreeCAD.Metadata]]:
        """Execute the dialog as a modal, returning a new Metadata object if the dialog
        is accepted, or None if it is rejected. This metadata object represents a single
        new content item. It's returned as a tuple with the object type as the first component,
        and the metadata object itself as the second."""
        if metadata:
            self.metadata = FreeCAD.Metadata(metadata)  # Deep copy
        else:
            self.metadata = FreeCAD.Metadata()
        self.dialog.singletonCheckBox.setChecked(singleton)
        if singleton:
            # This doesn't happen automatically the first time
            self.dialog.otherMetadataGroupBox.hide()
        index = self.dialog.addonKindComboBox.findData(content_kind)
        if index == -1:
            index = 2  # Workbench
            FreeCAD.Console.PrintWarning(
                translate("AddonsInstaller", "Unrecognized content kind '{}'").format(
                    content_kind
                )
                + "\n"
            )
        self.dialog.addonKindComboBox.setCurrentIndex(index)
        if metadata:
            self._populate_dialog(metadata)

        self.dialog.layout().setSizeConstraint(QLayout.SetFixedSize)
        result = self.dialog.exec()
        if result == QDialog.Accepted:
            return self._generate_metadata()
        return None

    def _populate_dialog(self, metadata: FreeCAD.Metadata) -> None:
        """Fill in the dialog with the details from the passed metadata object"""
        addon_kind = self.dialog.addonKindComboBox.currentData()
        if addon_kind == "workbench":
            self.dialog.workbenchClassnameLineEdit.setText(metadata.Classname)
        elif addon_kind == "macro":
            files = self.metadata.File
            if files:
                self.dialog.macroFileLineEdit.setText(files[0])
        elif addon_kind == "preferencepack":
            self.dialog.prefPackNameLineEdit.setText(self.metadata.Name)
        else:
            raise RuntimeError("Invalid data found for selection")

        # Now set the rest of it
        if metadata.Icon:
            self._set_icon(metadata.Icon)
        elif self.toplevel_metadata.Icon:
            if metadata.Subdirectory and metadata.Subdirectory != "./":
                self._set_icon("../" + self.toplevel_metadata.Icon)
            else:
                self._set_icon(self.toplevel_metadata.Icon)
        else:
            self.dialog.iconLabel.hide()
            self.dialog.iconLineEdit.setText("")

        if metadata.Subdirectory:
            self.dialog.subdirectoryLineEdit.setText(metadata.Subdirectory)
        else:
            self.dialog.subdirectoryLineEdit.setText("")

        self.dialog.displayNameLineEdit.setText(metadata.Name)
        self.dialog.descriptionTextEdit.setPlainText(metadata.Description)
        self.dialog.versionLineEdit.setText(metadata.Version)

        self.people_table.show(metadata)
        self.licenses_table.show(metadata, self.path_to_addon)

    def _set_icon(self, icon_relative_path):
        """Load the icon and display it, and its path, in the dialog."""
        icon_path = os.path.join(
            self.path_to_addon, icon_relative_path.replace("/", os.path.sep)
        )
        if os.path.isfile(icon_path):
            icon_data = QIcon(icon_path)
            if not icon_data.isNull():
                self.dialog.iconLabel.setPixmap(icon_data.pixmap(32, 32))
                self.dialog.iconLabel.show()
        else:
            FreeCAD.Console.PrintError(
                translate("AddonsInstaller", "Unable to locate icon at {}").format(
                    icon_path
                )
                + "\n"
            )
        self.dialog.iconLineEdit.setText(icon_relative_path)

    def _generate_metadata(self) -> Tuple[str, FreeCAD.Metadata]:
        """Create and return a new metadata object based on the contents of the dialog."""

        if not self.metadata:
            self.metadata = FreeCAD.Metadata()

        ##########################################################################################
        # Required data:
        current_data: str = self.dialog.addonKindComboBox.currentData()
        if current_data == "preferencepack":
            self.metadata.Name = self.dialog.prefPackNameLineEdit.text()
        elif self.dialog.displayNameLineEdit.text():
            self.metadata.Name = self.dialog.displayNameLineEdit.text()

        if current_data == "workbench":
            self.metadata.Classname = self.dialog.workbenchClassnameLineEdit.text()
        elif current_data == "macro":
            self.metadata.File = [self.dialog.macroFileLineEdit.text()]
        ##########################################################################################

        self.metadata.Subdirectory = self.dialog.subdirectoryLineEdit.text()
        self.metadata.Icon = self.dialog.iconLineEdit.text()

        # Early return if this is the only addon
        if self.dialog.singletonCheckBox.isChecked():
            return current_data, self.metadata

        # Otherwise, process the rest of the metadata (display name is already done)
        self.metadata.Description = (
            self.dialog.descriptionTextEdit.document().toPlainText()
        )
        self.metadata.Version = self.dialog.versionLineEdit.text()

        maintainers = []
        authors = []
        for row in range(self.dialog.peopleTableWidget.rowCount()):
            person_type = self.dialog.peopleTableWidget.item(row, 0).data()
            name = self.dialog.peopleTableWidget.item(row, 1).text()
            email = self.dialog.peopleTableWidget.item(row, 2).text()
            if person_type == "maintainer":
                maintainers.append({"name": name, "email": email})
            elif person_type == "author":
                authors.append({"name": name, "email": email})
        self.metadata.Maintainer = maintainers
        self.metadata.Author = authors

        licenses = []
        for row in range(self.dialog.licensesTableWidget.rowCount()):
            new_license = {
                "name": self.dialog.licensesTableWidget.item(row, 0).text,
                "file": self.dialog.licensesTableWidget.item(row, 1).text(),
            }
            licenses.append(new_license)
        self.metadata.License = licenses

        return self.dialog.addonKindComboBox.currentData(), self.metadata

    ###############################################################################################
    #                                         DIALOG SLOTS
    ###############################################################################################

    def _browse_for_icon_clicked(self):
        """Callback: when the "Browse..." button for the icon field is clicked"""
        subdir = self.dialog.subdirectoryLineEdit.text()
        start_dir = os.path.join(self.path_to_addon, subdir)
        new_icon_path, _ = QFileDialog.getOpenFileName(
            parent=self.dialog,
            caption=translate(
                "AddonsInstaller",
                "Select an icon file for this content item",
            ),
            dir=start_dir,
        )

        if not new_icon_path:
            return

        base_path = self.path_to_addon.replace("/", os.path.sep)
        icon_path = new_icon_path.replace("/", os.path.sep)
        if base_path[-1] != os.path.sep:
            base_path += os.path.sep

        if not icon_path.startswith(base_path):
            FreeCAD.Console.PrintError(
                translate("AddonsInstaller", "{} is not a subdirectory of {}").format(
                    icon_path, base_path
                )
                + "\n"
            )
            return
        self._set_icon(new_icon_path[len(base_path) :])
        self.metadata.Icon = new_icon_path[len(base_path) :]

    def _browse_for_subdirectory_clicked(self):
        """Callback: when the "Browse..." button for the subdirectory field is clicked"""
        subdir = self.dialog.subdirectoryLineEdit.text()
        start_dir = os.path.join(self.path_to_addon, subdir)
        new_subdir_path = QFileDialog.getExistingDirectory(
            parent=self.dialog,
            caption=translate(
                "AddonsInstaller",
                "Select the subdirectory for this content item",
            ),
            dir=start_dir,
        )
        if not new_subdir_path:
            return
        if new_subdir_path[-1] != "/":
            new_subdir_path += "/"

        # Three legal possibilities:
        #   1) This might be the toplevel directory, in which case we want to set
        #      metadata.Subdirectory to "./"
        #   2) This might be a subdirectory with the same name as the content item, in which case
        #      we don't need to set metadata.Subdirectory at all
        #   3) This might be some other directory name, but still contained within the top-level
        #      directory, in which case we want to set metadata.Subdirectory to the relative path

        # First, reject anything that isn't within the appropriate directory structure:
        base_path = self.path_to_addon.replace("/", os.path.sep)
        subdir_path = new_subdir_path.replace("/", os.path.sep)
        if not subdir_path.startswith(base_path):
            FreeCAD.Console.PrintError(
                translate("AddonsInstaller", "{} is not a subdirectory of {}").format(
                    subdir_path, base_path
                )
                + "\n"
            )
            return

        relative_path = subdir_path[len(base_path) :]
        if not relative_path:
            relative_path = "./"
        elif relative_path[-1] == os.path.sep:
            relative_path = relative_path[:-1]
        self.dialog.subdirectoryLineEdit.setText(relative_path)

    def _tags_clicked(self):
        """Show the tag editor"""
        tags = []
        if not self.metadata:
            self.metadata = FreeCAD.Metadata()
        if self.metadata:
            tags = self.metadata.Tag
        dlg = EditTags(tags)
        new_tags = dlg.exec()
        self.metadata.Tag = new_tags

    def _freecad_versions_clicked(self):
        """Show the FreeCAD version editor"""
        if not self.metadata:
            self.metadata = FreeCAD.Metadata()
        dlg = EditFreeCADVersions()
        dlg.exec(self.metadata)

    def _dependencies_clicked(self):
        """Show the dependencies editor"""
        if not self.metadata:
            self.metadata = FreeCAD.Metadata()
        dlg = EditDependencies()
        dlg.exec(self.metadata)  # Modifies metadata directly


class EditTags:
    """A dialog to edit tags"""

    def __init__(self, tags: List[str] = None):
        self.dialog = FreeCADGui.PySideUic.loadUi(
            os.path.join(os.path.dirname(__file__), "developer_mode_tags.ui")
        )
        self.original_tags = tags
        if tags:
            self.dialog.lineEdit.setText(", ".join(tags))

    def exec(self):
        """Execute the dialog, returning a list of tags (which may be empty, but still represents
        the expected list of tags to be set, e.g. the user may have removed them all).
        """
        result = self.dialog.exec()
        if result == QDialog.Accepted:
            new_tags: List[str] = self.dialog.lineEdit.text().split(",")
            clean_tags: List[str] = []
            for tag in new_tags:
                clean_tags.append(tag.strip())
            return clean_tags
        return self.original_tags


class EditDependencies:
    """A dialog to edit dependency information"""

    def __init__(self):
        self.dialog = FreeCADGui.PySideUic.loadUi(
            os.path.join(os.path.dirname(__file__), "developer_mode_dependencies.ui")
        )
        self.dialog.addDependencyToolButton.setIcon(
            QIcon.fromTheme("add", QIcon(":/icons/list-add.svg"))
        )
        self.dialog.removeDependencyToolButton.setIcon(
            QIcon.fromTheme("remove", QIcon(":/icons/list-remove.svg"))
        )
        self.dialog.addDependencyToolButton.clicked.connect(
            self._add_dependency_clicked
        )
        self.dialog.removeDependencyToolButton.clicked.connect(
            self._remove_dependency_clicked
        )
        self.dialog.tableWidget.itemDoubleClicked.connect(self._edit_dependency)
        self.dialog.tableWidget.itemSelectionChanged.connect(
            self._current_index_changed
        )

        self.dialog.removeDependencyToolButton.setDisabled(True)
        self.metadata = None

    def exec(self, metadata: FreeCAD.Metadata):
        """Execute the dialog"""
        self.metadata = FreeCAD.Metadata(metadata)  # Make a copy, in case we cancel
        row = 0
        for dep in self.metadata.Depend:
            dep_type = dep["type"]
            dep_name = dep["package"]
            dep_optional = dep["optional"]
            self._add_row(row, dep_type, dep_name, dep_optional)
            row += 1
        result = self.dialog.exec()
        if result == QDialog.Accepted:
            metadata.Depend = self.metadata.Depend

    def _add_dependency_clicked(self):
        """Callback: The add button was clicked"""
        dlg = EditDependency()
        dep_type, dep_name, dep_optional = dlg.exec()
        if dep_name:
            row = self.dialog.tableWidget.rowCount()
            self._add_row(row, dep_type, dep_name, dep_optional)
            self.metadata.addDepend(
                {"package": dep_name, "type": dep_type, "optional": dep_optional}
            )

    def _add_row(self, row, dep_type, dep_name, dep_optional):
        """Utility function to add a row to the table."""
        translations = {
            "automatic": translate("AddonsInstaller", "Automatic"),
            "workbench": translate("AddonsInstaller", "Workbench"),
            "addon": translate("AddonsInstaller", "Addon"),
            "python": translate("AddonsInstaller", "Python"),
        }
        if dep_type and dep_name:
            self.dialog.tableWidget.insertRow(row)
            type_item = QTableWidgetItem(translations[dep_type])
            type_item.setData(Qt.UserRole, dep_type)
            self.dialog.tableWidget.setItem(row, 0, type_item)
            self.dialog.tableWidget.setItem(row, 1, QTableWidgetItem(dep_name))
            if dep_optional:
                self.dialog.tableWidget.setItem(
                    row, 2, QTableWidgetItem(translate("AddonsInstaller", "Yes"))
                )

    def _remove_dependency_clicked(self):
        """Callback: The remove button was clicked"""
        items = self.dialog.tableWidget.selectedItems()
        if items:
            row = items[0].row()
            dep_type = self.dialog.tableWidget.item(row, 0).data(Qt.UserRole)
            dep_name = self.dialog.tableWidget.item(row, 1).text()
            dep_optional = bool(self.dialog.tableWidget.item(row, 2))
            self.metadata.removeDepend(
                {"package": dep_name, "type": dep_type, "optional": dep_optional}
            )
            self.dialog.tableWidget.removeRow(row)

    def _edit_dependency(self, item):
        """Callback: the dependency was double-clicked"""
        row = item.row()
        dlg = EditDependency()
        dep_type = self.dialog.tableWidget.item(row, 0).data(Qt.UserRole)
        dep_name = self.dialog.tableWidget.item(row, 1).text()
        dep_optional = bool(self.dialog.tableWidget.item(row, 2))
        new_dep_type, new_dep_name, new_dep_optional = dlg.exec(
            dep_type, dep_name, dep_optional
        )
        if dep_type and dep_name:
            self.metadata.removeDepend(
                {"package": dep_name, "type": dep_type, "optional": dep_optional}
            )
            self.metadata.addDepend(
                {
                    "package": new_dep_name,
                    "type": new_dep_type,
                    "optional": new_dep_optional,
                }
            )
            self.dialog.tableWidget.removeRow(row)
            self._add_row(row, dep_type, dep_name, dep_optional)

    def _current_index_changed(self):
        if self.dialog.tableWidget.selectedItems():
            self.dialog.removeDependencyToolButton.setDisabled(False)
        else:
            self.dialog.removeDependencyToolButton.setDisabled(True)


class EditDependency:
    """A dialog to edit a single piece of dependency information"""

    def __init__(self):
        self.dialog = FreeCADGui.PySideUic.loadUi(
            os.path.join(os.path.dirname(__file__), "developer_mode_edit_dependency.ui")
        )

        self.dialog.typeComboBox.addItem(
            translate("AddonsInstaller", "Internal Workbench"), "workbench"
        )
        self.dialog.typeComboBox.addItem(
            translate("AddonsInstaller", "External Addon"), "addon"
        )
        self.dialog.typeComboBox.addItem(
            translate("AddonsInstaller", "Python Package"), "python"
        )

        self.dialog.typeComboBox.currentIndexChanged.connect(
            self._type_selection_changed
        )
        self.dialog.dependencyComboBox.currentIndexChanged.connect(
            self._dependency_selection_changed
        )

        # Expect mostly Python dependencies...
        self.dialog.typeComboBox.setCurrentIndex(2)

        self.dialog.layout().setSizeConstraint(QLayout.SetFixedSize)

    def exec(
        self, dep_type="", dep_name="", dep_optional=False
    ) -> Tuple[str, str, bool]:
        """Execute the dialog, returning a tuple of the type of dependency (workbench, addon, or
        python), the name of the dependency, and a boolean indicating whether this is optional.
        """

        # If we are editing an existing row, set up the dialog:
        if dep_type and dep_name:
            index = self.dialog.typeComboBox.findData(dep_type)
            if index == -1:
                raise RuntimeError(f"Invalid dependency type {dep_type}")
            self.dialog.typeComboBox.setCurrentIndex(index)
            index = self.dialog.dependencyComboBox.findData(dep_name)
            if index == -1:
                index = self.dialog.dependencyComboBox.findData("other")
            self.dialog.dependencyComboBox.setCurrentIndex(index)
            self.dialog.lineEdit.setText(dep_name)
            self.dialog.optionalCheckBox.setChecked(dep_optional)

        # Run the dialog (modal)
        result = self.dialog.exec()
        if result == QDialog.Accepted:
            dep_type = self.dialog.typeComboBox.currentData()
            dep_optional = self.dialog.optionalCheckBox.isChecked()
            dep_name = self.dialog.dependencyComboBox.currentData()
            if dep_name == "other":
                dep_name = self.dialog.lineEdit.text()
            return dep_type, dep_name, dep_optional
        return "", "", False

    def _populate_internal_workbenches(self):
        """Add all known internal FreeCAD Workbenches to the list"""
        self.dialog.dependencyComboBox.clear()
        for display_name, name in INTERNAL_WORKBENCHES.items():
            self.dialog.dependencyComboBox.addItem(display_name, name)
        # No "other" option is supported for this type of dependency

    def _populate_external_addons(self):
        """Add all known addons to the list"""
        self.dialog.dependencyComboBox.clear()
        # pylint: disable=import-outside-toplevel
        from AddonManager import INSTANCE as AM_INSTANCE

        repo_dict = {}
        # We need a case-insensitive sorting of all repo types, displayed and sorted by their
        # display name, but keeping track of their official name as well (stored in the UserRole)
        for repo in AM_INSTANCE.item_model.repos:
            repo_dict[repo.display_name.lower()] = (repo.display_name, repo.name)
        sorted_keys = sorted(repo_dict.keys())
        for item in sorted_keys:
            self.dialog.dependencyComboBox.addItem(
                repo_dict[item][0], repo_dict[item][1]
            )
        self.dialog.dependencyComboBox.addItem(
            translate("AddonsInstaller", "Other..."), "other"
        )

    def _populate_allowed_python_packages(self):
        """Add all allowed python packages to the list"""
        self.dialog.dependencyComboBox.clear()
        # pylint: disable=import-outside-toplevel
        from AddonManager import INSTANCE as AM_INSTANCE

        packages = sorted(AM_INSTANCE.allowed_packages)
        for package in packages:
            self.dialog.dependencyComboBox.addItem(package, package)
        self.dialog.dependencyComboBox.addItem(
            translate("AddonsInstaller", "Other..."), "other"
        )

    def _type_selection_changed(self, _):
        """Callback: The type of dependency has been changed"""
        selection = self.dialog.typeComboBox.currentData()
        if selection == "workbench":
            self._populate_internal_workbenches()
        elif selection == "addon":
            self._populate_external_addons()
        elif selection == "python":
            self._populate_allowed_python_packages()
        else:
            raise RuntimeError("Invalid data found for selection")

    def _dependency_selection_changed(self, _):
        selection = self.dialog.dependencyComboBox.currentData()
        if selection == "other":
            self.dialog.lineEdit.show()
            self.dialog.otherNote.show()
        else:
            self.dialog.lineEdit.hide()
            self.dialog.otherNote.hide()


class EditFreeCADVersions:
    """A dialog to edit minimum and maximum FreeCAD version support"""

    def __init__(self):
        self.dialog = FreeCADGui.PySideUic.loadUi(
            os.path.join(
                os.path.dirname(__file__), "developer_mode_freecad_versions.ui"
            )
        )

    def exec(self, metadata: FreeCAD.Metadata):
        """Execute the dialog"""
        if metadata.FreeCADMin != "0.0.0":
            self.dialog.minVersionLineEdit.setText(metadata.FreeCADMin)
        if metadata.FreeCADMax != "0.0.0":
            self.dialog.maxVersionLineEdit.setText(metadata.FreeCADMax)
        result = self.dialog.exec()
        if result == QDialog.Accepted:
            if self.dialog.minVersionLineEdit.text():
                metadata.FreeCADMin = self.dialog.minVersionLineEdit.text()
            else:
                metadata.FreeCADMin = None
            if self.dialog.maxVersionLineEdit.text():
                metadata.FreeCADMax = self.dialog.maxVersionLineEdit.text()
            else:
                metadata.FreeCADMax = None


class EditAdvancedVersions:
    """A dialog to support mapping specific git branches, tags, or commits to specific
    versions of FreeCAD."""

    def __init__(self):
        self.dialog = FreeCADGui.PySideUic.loadUi(
            os.path.join(
                os.path.dirname(__file__), "developer_mode_advanced_freecad_versions.ui"
            )
        )

    def exec(self):
        """Execute the dialog"""
        self.dialog.exec()



#***************************************************************************
#*                                                                         *
#*   Copyright (c) 2021 Yorik van Havre <yorik@uncreated.net>              *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************


"""This module contains FreeCAD commands for the Draft workbench"""

import os
import FreeCAD
import draftguitools.gui_base as gui_base

from draftutils.translate import translate, QT_TRANSLATE_NOOP

class Draft_Hatch(gui_base.GuiCommandSimplest):


    def GetResources(self):

        return {'Pixmap'  : "Draft_Hatch",
                'MenuText': QT_TRANSLATE_NOOP("Draft_Hatch", "Hatch"),
                'Accel': "H, A",
                'ToolTip' : QT_TRANSLATE_NOOP("Draft_Hatch", "Creates hatches on the faces of a selected object")}

    def Activated(self):

        import FreeCADGui

        if FreeCADGui.Selection.getSelection():
            FreeCADGui.Control.showDialog(Draft_Hatch_TaskPanel(FreeCADGui.Selection.getSelection()[0]))
        else:
            FreeCAD.Console.PrintError(translate("Draft","You must choose a base object before using this command")+"\n")


class Draft_Hatch_TaskPanel:


    def __init__(self,baseobj):

        import FreeCADGui
        from PySide import QtCore,QtGui
        import Draft_rc

        self.baseobj = baseobj
        self.form = FreeCADGui.PySideUic.loadUi(":/ui/dialogHatch.ui")
        self.form.setWindowIcon(QtGui.QIcon(":/icons/Draft_Hatch.svg"))
        self.form.File.fileNameChanged.connect(self.onFileChanged)
        self.p1 = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/TechDraw/PAT")
        self.p2 = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft")
        self.form.File.setFileName(self.p1.GetString("FilePattern",""))
        pat = self.p1.GetString("NamePattern","")
        if pat in [self.form.Pattern.itemText(i) for i in range(self.form.Pattern.count())]:
            self.form.Pattern.setCurrentText(pat)
        self.form.Scale.setValue(self.p2.GetFloat("HatchPatternScale",1000.0))
        self.form.Rotation.setValue(self.p2.GetFloat("HatchPatternRotation",0.0))

    def accept(self):

        import FreeCADGui

        self.p1.SetString("FilePattern",self.form.File.property("fileName"))
        self.p1.SetString("NamePattern",self.form.Pattern.currentText())
        self.p2.SetFloat("HatchPatternScale",self.form.Scale.value())
        self.p2.SetFloat("HatchPatternRotation",self.form.Rotation.value())
        if hasattr(self.baseobj,"File") and hasattr(self.baseobj,"Pattern"):
            # modify existing hatch object
            o = "FreeCAD.ActiveDocument.getObject(\""+self.baseobj.Name+"\")"
            FreeCADGui.doCommand(o+".File=\""+self.form.File.property("fileName")+"\"")
            FreeCADGui.doCommand(o+".Pattern=\""+self.form.Pattern.currentText()+"\"")
            FreeCADGui.doCommand(o+".Scale="+str(self.form.Scale.value()))
            FreeCADGui.doCommand(o+".Rotation="+str(self.form.Rotation.value()))
        else:
            # create new hatch object
            FreeCAD.ActiveDocument.openTransaction("Create Hatch")
            FreeCADGui.addModule("Draft")
            cmd = "Draft.make_hatch("
            cmd += "baseobject=FreeCAD.ActiveDocument.getObject(\""+self.baseobj.Name
            cmd += "\"),filename=\""+self.form.File.property("fileName")
            cmd += "\",pattern=\""+self.form.Pattern.currentText()
            cmd += "\",scale="+str(self.form.Scale.value())
            cmd += ",rotation="+str(self.form.Rotation.value())+")"
            FreeCADGui.doCommand(cmd)
            FreeCAD.ActiveDocument.commitTransaction()
        FreeCADGui.doCommand("FreeCAD.ActiveDocument.recompute()")
        self.reject()

    def reject(self):

        import FreeCADGui

        FreeCADGui.Control.closeDialog()
        FreeCADGui.ActiveDocument.resetEdit()
        FreeCAD.ActiveDocument.recompute()

    def onFileChanged(self,filename):

        pat = self.form.Pattern.currentText()
        self.form.Pattern.clear()
        patterns = self.getPatterns(filename)
        self.form.Pattern.addItems(patterns)
        if pat in patterns:
            self.form.Pattern.setCurrentText(pat)

    def getPatterns(self,filename):

        """returns a list of pattern names found in a PAT file"""
        patterns = []
        if os.path.exists(filename):
            with open(filename) as patfile:
                for line in patfile:
                    if line.startswith("*"):
                        patterns.append(line.split(",")[0][1:])
        return patterns

if FreeCAD.GuiUp:
    import FreeCADGui
    FreeCADGui.addCommand("Draft_Hatch",Draft_Hatch())



# ***************************************************************************
# *   Copyright (c) 2018 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import os
import FreeCAD
import FreeCADGui

from urllib.parse import unquote


# filename will be given before this script is run
cfolders = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Start").GetString(
    "ShowCustomFolder", ""
)
if cfolders:
    dirnumber = int(filename[0])
    filename = filename[2:]
    cfolder = cfolders.split(";;")[dirnumber]
    if not os.path.isdir(cfolder):
        cfolder = os.path.dirname(cfolder)
    f = unquote(filename).replace("+", " ")
    ext = os.path.splitext(filename)[1].lower().strip(".")
    mod = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Start").GetString(
        "DefaultImport" + ext, ""
    )
    if (
        ext.lower() in ("fcstd", "stp", "step", "iges", "igs")
    ):
        FreeCAD.loadFile(os.path.join(cfolder, f), mod)
    elif (
        ext.lower() in ("bmp", "cur", "gif", "ico", "pbm", "pgm", "png", "jpg",
                        "jpeg", "ppm", "svg", "svgz", "xbm", "xpm")
    ):
        FreeCAD.newDocument()
        FreeCADGui.insert(os.path.join(cfolder, f), FreeCAD.activeDocument().Name)
        FreeCAD.activeDocument().recompute()
    else:
        FreeCADGui.loadFile(os.path.join(cfolder, f), mod)
    FreeCADGui.activeDocument().sendMsgToViews("ViewFit")

    from StartPage import StartPage
    StartPage.postStart()



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022-2023 FreeCAD Project Association                   *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Provides classes and support functions for managing the automatically-installed
Python library dependencies. No support is provided for uninstalling those dependencies
because pip's uninstall function does not support the target directory argument. """

import json
import os
import platform
import shutil
import subprocess
import sys
from functools import partial
from typing import Dict, List, Tuple

import FreeCAD
import FreeCADGui
from PySide import QtCore, QtGui, QtWidgets

import addonmanager_utilities as utils

translate = FreeCAD.Qt.translate

# pylint: disable=too-few-public-methods


class PipFailed(Exception):
    """Exception thrown when pip times out or otherwise fails to return valid results"""


class CheckForPythonPackageUpdatesWorker(QtCore.QThread):
    """Perform non-blocking Python library update availability checking"""

    python_package_updates_available = QtCore.Signal()

    def __init__(self):
        QtCore.QThread.__init__(self)

    def run(self):
        """Usually not called directly: instead, instantiate this class and call its start()
        function in a parent thread. emits a python_package_updates_available signal if updates
        are available for any of the installed Python packages."""

        if check_for_python_package_updates():
            self.python_package_updates_available.emit()


def check_for_python_package_updates() -> bool:
    """Returns True if any of the Python packages installed into the AdditionalPythonPackages
    directory have updates available, or False if the are all up-to-date."""

    vendor_path = os.path.join(FreeCAD.getUserAppDataDir(), "AdditionalPythonPackages")
    package_counter = 0
    try:
        outdated_packages_stdout = call_pip(["list", "-o", "--path", vendor_path])
    except PipFailed as e:
        FreeCAD.Console.PrintError(str(e) + "\n")
        return False
    FreeCAD.Console.PrintLog("Output from pip -o:\n")
    for line in outdated_packages_stdout:
        if len(line) > 0:
            package_counter += 1
        FreeCAD.Console.PrintLog(f"  {line}\n")
    return package_counter > 0


def call_pip(args) -> List[str]:
    """Tries to locate the appropriate Python executable and run pip with version checking
    disabled. Fails if Python can't be found or if pip is not installed."""

    python_exe = utils.get_python_exe()
    pip_failed = False
    if python_exe:
        call_args = [python_exe, "-m", "pip", "--disable-pip-version-check"]
        call_args.extend(args)
        proc = None
        try:
            proc = utils.run_interruptable_subprocess(call_args)
        except subprocess.CalledProcessError:
            pip_failed = True

        result = []
        if not pip_failed:
            data = proc.stdout
            result = data.split("\n")
        elif proc:
            raise PipFailed(proc.stderr)
        else:
            raise PipFailed("pip timed out")
    else:
        raise PipFailed("Could not locate Python executable on this system")
    return result


class PythonPackageManager:

    """A GUI-based pip interface allowing packages to be updated, either individually or all at
    once."""

    class PipRunner(QtCore.QObject):
        """Run pip in a separate thread so the UI doesn't block while it runs"""

        finished = QtCore.Signal()
        error = QtCore.Signal(str)

        def __init__(self, vendor_path, parent=None):
            super().__init__(parent)
            self.all_packages_stdout = []
            self.outdated_packages_stdout = []
            self.vendor_path = vendor_path
            self.package_list = {}

        def process(self):
            """Execute this object."""
            try:
                self.all_packages_stdout = call_pip(
                    ["list", "--path", self.vendor_path]
                )
                self.outdated_packages_stdout = call_pip(
                    ["list", "-o", "--path", self.vendor_path]
                )
            except PipFailed as e:
                FreeCAD.Console.PrintError(str(e) + "\n")
                self.error.emit(str(e))
            self.finished.emit()

    def __init__(self, addons):
        self.dlg = FreeCADGui.PySideUic.loadUi(
            os.path.join(os.path.dirname(__file__), "PythonDependencyUpdateDialog.ui")
        )
        self.addons = addons
        self.vendor_path = utils.get_pip_target_directory()
        self.worker_thread = None
        self.worker_object = None
        self.package_list = []

    def show(self):
        """Run the modal dialog"""

        known_python_versions = self.get_known_python_versions()
        if self._current_python_version_is_new() and known_python_versions:
            # pylint: disable=line-too-long
            result = QtWidgets.QMessageBox.question(
                None,
                translate("AddonsInstaller", "New Python Version Detected"),
                translate(
                    "AddonsInstaller",
                    "This appears to be the first time this version of Python has been used with the Addon Manager. Would you like to install the same auto-installed dependencies for it?",
                ),
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            )
            if result == QtWidgets.QMessageBox.Yes:
                self._reinstall_all_packages()

        self._add_current_python_version()
        self._create_list_from_pip()
        self.dlg.setWindowFlag(QtCore.Qt.WindowStaysOnTopHint, True)
        self.dlg.tableWidget.setSortingEnabled(False)
        self.dlg.labelInstallationPath.setText(self.vendor_path)
        self.dlg.exec()

    def _create_list_from_pip(self):
        """Uses pip and pip -o to generate a list of installed packages, and creates the user
        interface elements for those packages. Asynchronous, will complete AFTER the window is
        showing in most cases."""

        self.worker_thread = QtCore.QThread()
        self.worker_object = PythonPackageManager.PipRunner(self.vendor_path)
        self.worker_object.moveToThread(self.worker_thread)
        self.worker_object.finished.connect(self._worker_finished)
        self.worker_object.finished.connect(self.worker_thread.quit)
        self.worker_thread.started.connect(self.worker_object.process)
        self.worker_thread.start()

        self.dlg.tableWidget.setRowCount(1)
        self.dlg.tableWidget.setItem(
            0,
            0,
            QtWidgets.QTableWidgetItem(
                translate("AddonsInstaller", "Processing, please wait...")
            ),
        )
        self.dlg.tableWidget.horizontalHeader().setSectionResizeMode(
            0, QtWidgets.QHeaderView.ResizeToContents
        )

    def _worker_finished(self):
        """Callback for when the worker process has completed"""
        all_packages_stdout = self.worker_object.all_packages_stdout
        outdated_packages_stdout = self.worker_object.outdated_packages_stdout

        self.package_list = self._parse_pip_list_output(
            all_packages_stdout, outdated_packages_stdout
        )
        self.dlg.buttonUpdateAll.clicked.connect(
            partial(self._update_all_packages, self.package_list)
        )

        self.dlg.tableWidget.setRowCount(len(self.package_list))
        updateButtons = []
        counter = 0
        update_counter = 0
        self.dlg.tableWidget.setSortingEnabled(False)
        for package_name, package_details in self.package_list.items():
            dependent_addons = self._get_dependent_addons(package_name)
            dependencies = []
            for addon in dependent_addons:
                if addon["optional"]:
                    dependencies.append(addon["name"] + "*")
                else:
                    dependencies.append(addon["name"])
            self.dlg.tableWidget.setItem(
                counter, 0, QtWidgets.QTableWidgetItem(package_name)
            )
            self.dlg.tableWidget.setItem(
                counter,
                1,
                QtWidgets.QTableWidgetItem(package_details["installed_version"]),
            )
            self.dlg.tableWidget.setItem(
                counter,
                2,
                QtWidgets.QTableWidgetItem(package_details["available_version"]),
            )
            self.dlg.tableWidget.setItem(
                counter,
                3,
                QtWidgets.QTableWidgetItem(", ".join(dependencies)),
            )
            if len(package_details["available_version"]) > 0:
                updateButtons.append(
                    QtWidgets.QPushButton(translate("AddonsInstaller", "Update"))
                )
                updateButtons[-1].setIcon(QtGui.QIcon(":/icons/button_up.svg"))
                updateButtons[-1].clicked.connect(
                    partial(self._update_package, package_name)
                )
                self.dlg.tableWidget.setCellWidget(counter, 4, updateButtons[-1])
                update_counter += 1
            else:
                self.dlg.tableWidget.removeCellWidget(counter, 3)
            counter += 1
        self.dlg.tableWidget.setSortingEnabled(True)

        self.dlg.tableWidget.horizontalHeader().setStretchLastSection(False)
        self.dlg.tableWidget.horizontalHeader().setSectionResizeMode(
            0, QtWidgets.QHeaderView.Stretch
        )
        self.dlg.tableWidget.horizontalHeader().setSectionResizeMode(
            1, QtWidgets.QHeaderView.ResizeToContents
        )
        self.dlg.tableWidget.horizontalHeader().setSectionResizeMode(
            2, QtWidgets.QHeaderView.ResizeToContents
        )
        self.dlg.tableWidget.horizontalHeader().setSectionResizeMode(
            3, QtWidgets.QHeaderView.ResizeToContents
        )

        if update_counter > 0:
            self.dlg.buttonUpdateAll.setEnabled(True)
        else:
            self.dlg.buttonUpdateAll.setEnabled(False)

    def _get_dependent_addons(self, package):
        dependent_addons = []
        for addon in self.addons:
            # if addon.installed_version is not None:
            if package.lower() in addon.python_requires:
                dependent_addons.append({"name": addon.name, "optional": False})
            elif package.lower() in addon.python_optional:
                dependent_addons.append({"name": addon.name, "optional": True})
        return dependent_addons

    def _parse_pip_list_output(
        self, all_packages, outdated_packages
    ) -> Dict[str, Dict[str, str]]:
        """Parses the output from pip into a dictionary with update information in it. The pip
        output should be an array of lines of text."""

        # All Packages output looks like this:
        # Package    Version
        # ---------- -------
        # gitdb      4.0.9
        # GitPython  3.1.27
        # setuptools 41.2.0

        # Outdated Packages output looks like this:
        # Package    Version Latest Type
        # ---------- ------- ------ -----
        # pip        21.0.1  22.1.2 wheel
        # setuptools 41.2.0  63.2.0 wheel

        packages = {}
        skip_counter = 0
        for line in all_packages:
            if skip_counter < 2:
                skip_counter += 1
                continue
            entries = line.split()
            if len(entries) > 1:
                package_name = entries[0]
                installed_version = entries[1]
                packages[package_name] = {
                    "installed_version": installed_version,
                    "available_version": "",
                }

        skip_counter = 0
        for line in outdated_packages:
            if skip_counter < 2:
                skip_counter += 1
                continue
            entries = line.split()
            if len(entries) > 1:
                package_name = entries[0]
                installed_version = entries[1]
                available_version = entries[2]
                packages[package_name] = {
                    "installed_version": installed_version,
                    "available_version": available_version,
                }

        return packages

    def _update_package(self, package_name) -> None:
        """Run pip --upgrade on the given package. Updates all dependent packages as well."""
        for line in range(self.dlg.tableWidget.rowCount()):
            if self.dlg.tableWidget.item(line, 0).text() == package_name:
                self.dlg.tableWidget.setItem(
                    line,
                    2,
                    QtWidgets.QTableWidgetItem(
                        translate("AddonsInstaller", "Updating...")
                    ),
                )
                break
        QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents, 50)

        try:
            call_pip(
                ["install", "--upgrade", package_name, "--target", self.vendor_path]
            )
            self._create_list_from_pip()
        except PipFailed as e:
            FreeCAD.Console.PrintError(str(e) + "\n")
            return
        QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents, 50)

    def _update_all_packages(self, package_list) -> None:
        """Updates all packages with available updates."""
        updates = []
        for package_name, package_details in package_list.items():
            if (
                len(package_details["available_version"]) > 0
                and package_details["available_version"]
                != package_details["installed_version"]
            ):
                updates.append(package_name)

        for package_name in updates:
            self._update_package(package_name)

    @classmethod
    def migrate_old_am_installations(cls) -> bool:
        """Move packages installed before the Addon Manager switched to a versioned directory
        structure into the versioned structure. Returns True if a migration was done, or false
        if no migration was needed."""

        migrated = False

        old_directory = os.path.join(
            FreeCAD.getUserAppDataDir(), "AdditionalPythonPackages"
        )

        new_directory = utils.get_pip_target_directory()
        new_directory_name = new_directory.rsplit(os.path.sep, 1)[1]

        if not os.path.exists(old_directory) or os.path.exists(
            os.path.join(old_directory, "MIGRATION_COMPLETE")
        ):
            # Nothing to migrate
            return False

        if not os.path.exists(new_directory):
            os.makedirs(new_directory)

        for content_item in os.listdir(old_directory):
            if content_item == new_directory_name:
                continue
            old_path = os.path.join(old_directory, content_item)
            new_path = os.path.join(new_directory, content_item)
            FreeCAD.Console.PrintLog(
                f"Moving {content_item} into the new (versioned) directory structure\n"
            )
            FreeCAD.Console.PrintLog(f"   {old_path} --> {new_path}\n")
            shutil.move(old_path, new_path)
            migrated = True

        sys.path.append(new_directory)
        cls._add_current_python_version()

        with open(
            os.path.join(old_directory, "MIGRATION_COMPLETE"), "w", encoding="utf-8"
        ) as f:
            f.write(
                "Files originally installed in this directory have been migrated to:\n"
            )
            f.write(new_directory)
            f.write(
                "\nThe existence of this file prevents the Addon Manager from "
                "attempting the migration again.\n"
            )
        return migrated

    @classmethod
    def get_known_python_versions(cls) -> List[Tuple[int, int, int]]:
        """Get the list of Python versions that the Addon Manager has seen before."""
        pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")
        known_python_versions_string = pref.GetString("KnownPythonVersions", "[]")
        known_python_versions = json.loads(known_python_versions_string)
        print(json.dumps(known_python_versions, indent=4))
        return known_python_versions

    @classmethod
    def _add_current_python_version(cls) -> None:
        known_python_versions = cls.get_known_python_versions()
        major, minor, _ = platform.python_version_tuple()
        if not [major, minor] in known_python_versions:
            known_python_versions.append((major, minor))
        pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")
        pref.SetString("KnownPythonVersions", json.dumps(known_python_versions))

    @classmethod
    def _current_python_version_is_new(cls) -> bool:
        """Returns True if this is the first time the Addon Manager has seen this version of
        Python"""
        known_python_versions = cls.get_known_python_versions()
        major, minor, _ = platform.python_version_tuple()
        if not [major, minor] in known_python_versions:
            return True
        return False

    def _load_old_package_list(self) -> List[str]:
        """Gets the list of packages from the package installation manifest"""

        known_python_versions = self.get_known_python_versions()
        if not known_python_versions:
            return []
        last_version = known_python_versions[-1]
        expected_directory = f"py{last_version[0]}{last_version[1]}"
        expected_directory = os.path.join(
            FreeCAD.getUserAppDataDir(), "AdditionalPythonPackages", expected_directory
        )
        # For now just do this synchronously
        worker_object = PythonPackageManager.PipRunner(expected_directory)
        worker_object.process()
        packages = self._parse_pip_list_output(
            worker_object.all_packages_stdout, worker_object.outdated_packages_stdout
        )
        return packages.keys()

    def _reinstall_all_packages(self) -> None:
        """Loads the package manifest from another Python version, and installs the same packages
        for the current (presumably new) version of Python."""

        packages = self._load_old_package_list()
        args = ["install"]
        args.extend(packages)
        args.extend(["--target", self.vendor_path])

        try:
            call_pip(args)
        except PipFailed as e:
            FreeCAD.Console.PrintError(str(e) + "\n")
            return



#!/usr/bin/env python3

# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022-2023 FreeCAD Project Association                   *
# *   Copyright (c) 2015 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import os
import functools
import tempfile
import hashlib
import threading
import json
import re  # Needed for py 3.6 and earlier, can remove later, search for "re."
from datetime import date, timedelta
from typing import Dict

from PySide import QtGui, QtCore, QtWidgets
import FreeCAD
import FreeCADGui

from addonmanager_workers_startup import (
    CreateAddonListWorker,
    LoadPackagesFromCacheWorker,
    LoadMacrosFromCacheWorker,
    CheckWorkbenchesForUpdatesWorker,
    CacheMacroCodeWorker,
)
from addonmanager_workers_installation import (
    UpdateMetadataCacheWorker,
)
from addonmanager_installer_gui import AddonInstallerGUI, MacroInstallerGUI
from addonmanager_uninstaller_gui import AddonUninstallerGUI
from addonmanager_update_all_gui import UpdateAllGUI
import addonmanager_utilities as utils
import AddonManager_rc  # This is required by Qt, it's not unused
from package_list import PackageList, PackageListItemModel
from package_details import PackageDetails
from Addon import Addon
from manage_python_dependencies import (
    PythonPackageManager,
)
from addonmanager_devmode import DeveloperMode
from addonmanager_firstrun import FirstRunDialog
from addonmanager_connection_checker import ConnectionCheckerGUI
from addonmanager_devmode_metadata_checker import MetadataValidators

import NetworkManager

from AddonManagerOptions import AddonManagerOptions

translate = FreeCAD.Qt.translate


def QT_TRANSLATE_NOOP(_, txt):
    return txt


__title__ = "FreeCAD Addon Manager Module"
__author__ = "Yorik van Havre", "Jonathan Wiedemann", "Kurt Kremitzki", "Chris Hennes"
__url__ = "http://www.freecad.org"

"""
FreeCAD Addon Manager Module

Fetches various types of addons from a variety of sources. Built-in sources are:
* https://github.com/FreeCAD/FreeCAD-addons
* https://github.com/FreeCAD/FreeCAD-macros
* https://wiki.freecad.org/

Additional git sources may be configure via user preferences.

You need a working internet connection, and optionally git -- if git is not available, ZIP archives
are downloaded instead.
"""

#  \defgroup ADDONMANAGER AddonManager
#  \ingroup ADDONMANAGER
#  \brief The Addon Manager allows users to install workbenches and macros made by other users
#  @{

INSTANCE = None


class CommandAddonManager:
    """The main Addon Manager class and FreeCAD command"""

    workers = [
        "create_addon_list_worker",
        "check_worker",
        "show_worker",
        "showmacro_worker",
        "macro_worker",
        "update_metadata_cache_worker",
        "load_macro_metadata_worker",
        "update_all_worker",
        "check_for_python_package_updates_worker",
    ]

    lock = threading.Lock()
    restart_required = False

    def __init__(self):
        QT_TRANSLATE_NOOP("QObject", "Addon Manager")
        FreeCADGui.addPreferencePage(
            AddonManagerOptions,
            "Addon Manager",
        )

        self.check_worker = None
        self.check_for_python_package_updates_worker = None
        self.update_all_worker = None
        self.developer_mode = None
        self.installer_gui = None

        self.update_cache = False
        self.dialog = None
        self.startup_sequence = []

        # Set up the connection checker
        self.connection_checker = ConnectionCheckerGUI()
        self.connection_checker.connection_available.connect(self.launch)

        # Give other parts of the AM access to the current instance
        global INSTANCE
        INSTANCE = self

    def GetResources(self) -> Dict[str, str]:
        """FreeCAD-required function: get the core resource information for this Mod."""
        return {
            "Pixmap": "AddonManager",
            "MenuText": QT_TRANSLATE_NOOP("Std_AddonMgr", "&Addon manager"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Std_AddonMgr",
                "Manage external workbenches, macros, and preference packs",
            ),
            "Group": "Tools",
        }

    def Activated(self) -> None:
        """FreeCAD-required function: called when the command is activated."""
        NetworkManager.InitializeNetworkManager()
        firstRunDialog = FirstRunDialog()
        if not firstRunDialog.exec():
            return
        self.connection_checker.start()

    def launch(self) -> None:
        """Shows the Addon Manager UI"""

        # create the dialog
        self.dialog = FreeCADGui.PySideUic.loadUi(
            os.path.join(os.path.dirname(__file__), "AddonManager.ui")
        )
        self.dialog.setObjectName("AddonManager_Main_Window")
        # self.dialog.setWindowFlag(QtCore.Qt.WindowStaysOnTopHint, True)

        # cleanup the leftovers from previous runs
        self.macro_repo_dir = FreeCAD.getUserMacroDir(True)
        self.packages_with_updates = set()
        self.startup_sequence = []
        self.cleanup_workers()
        self.determine_cache_update_status()

        # restore window geometry from stored state
        pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")
        w = pref.GetInt("WindowWidth", 800)
        h = pref.GetInt("WindowHeight", 600)
        self.dialog.resize(w, h)

        # If we are checking for updates automatically, hide the Check for updates button:
        autocheck = pref.GetBool("AutoCheck", False)
        if autocheck:
            self.dialog.buttonCheckForUpdates.hide()
        else:
            self.dialog.buttonUpdateAll.hide()

        # Set up the listing of packages using the model-view-controller architecture
        self.packageList = PackageList(self.dialog)
        self.item_model = PackageListItemModel()
        self.packageList.setModel(self.item_model)
        self.dialog.contentPlaceholder.hide()
        self.dialog.layout().replaceWidget(self.dialog.contentPlaceholder, self.packageList)
        self.packageList.show()

        # Package details start out hidden
        self.packageDetails = PackageDetails(self.dialog)
        self.packageDetails.hide()
        index = self.dialog.layout().indexOf(self.packageList)
        self.dialog.layout().insertWidget(index, self.packageDetails)

        # set nice icons to everything, by theme with fallback to FreeCAD icons
        self.dialog.setWindowIcon(QtGui.QIcon(":/icons/AddonManager.svg"))
        self.dialog.buttonUpdateAll.setIcon(QtGui.QIcon(":/icons/button_valid.svg"))
        self.dialog.buttonCheckForUpdates.setIcon(QtGui.QIcon(":/icons/view-refresh.svg"))
        self.dialog.buttonClose.setIcon(
            QtGui.QIcon.fromTheme("close", QtGui.QIcon(":/icons/process-stop.svg"))
        )
        self.dialog.buttonPauseUpdate.setIcon(
            QtGui.QIcon.fromTheme("pause", QtGui.QIcon(":/icons/media-playback-stop.svg"))
        )

        pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")
        dev_mode_active = pref.GetBool("developerMode", False)

        # enable/disable stuff
        self.dialog.buttonUpdateAll.setEnabled(False)
        self.hide_progress_widgets()
        self.dialog.buttonUpdateCache.setEnabled(False)
        self.dialog.buttonUpdateCache.setText(translate("AddonsInstaller", "Starting up..."))
        if dev_mode_active:
            self.dialog.buttonDevTools.show()
        else:
            self.dialog.buttonDevTools.hide()

        # connect slots
        self.dialog.rejected.connect(self.reject)
        self.dialog.buttonUpdateAll.clicked.connect(self.update_all)
        self.dialog.buttonClose.clicked.connect(self.dialog.reject)
        self.dialog.buttonUpdateCache.clicked.connect(self.on_buttonUpdateCache_clicked)
        self.dialog.buttonPauseUpdate.clicked.connect(self.stop_update)
        self.dialog.buttonCheckForUpdates.clicked.connect(
            lambda: self.force_check_updates(standalone=True)
        )
        self.dialog.buttonUpdateDependencies.clicked.connect(self.show_python_updates_dialog)
        self.dialog.buttonDevTools.clicked.connect(self.show_developer_tools)
        self.packageList.itemSelected.connect(self.table_row_activated)
        self.packageList.setEnabled(False)
        self.packageDetails.execute.connect(self.executemacro)
        self.packageDetails.install.connect(self.launch_installer_gui)
        self.packageDetails.uninstall.connect(self.remove)
        self.packageDetails.update.connect(self.update)
        self.packageDetails.back.connect(self.on_buttonBack_clicked)
        self.packageDetails.update_status.connect(self.status_updated)

        # center the dialog over the FreeCAD window
        mw = FreeCADGui.getMainWindow()
        self.dialog.move(
            mw.frameGeometry().topLeft() + mw.rect().center() - self.dialog.rect().center()
        )

        # set info for the progress bar:
        self.dialog.progressBar.setMaximum(1000)

        # begin populating the table in a set of sub-threads
        self.startup()

        # set the label text to start with
        self.show_information(translate("AddonsInstaller", "Loading addon information"))

        # rock 'n roll!!!
        self.dialog.exec()

    def cleanup_workers(self) -> None:
        """Ensure that no workers are running by explicitly asking them to stop and waiting for
        them until they do"""
        for worker in self.workers:
            if hasattr(self, worker):
                thread = getattr(self, worker)
                if thread:
                    if not thread.isFinished():
                        thread.blockSignals(True)
                        thread.requestInterruption()
        for worker in self.workers:
            if hasattr(self, worker):
                thread = getattr(self, worker)
                if thread:
                    if not thread.isFinished():
                        finished = thread.wait(500)
                        if not finished:
                            FreeCAD.Console.PrintWarning(
                                translate(
                                    "AddonsInstaller",
                                    "Worker process {} is taking a long time to stop...",
                                ).format(worker)
                                + "\n"
                            )

    def determine_cache_update_status(self) -> None:
        """Determine whether we need to update the cache, based on user preference, and previous
        cache update status. Sets self.update_cache to either True or False."""

        # Figure out our cache update frequency: there is a combo box in the preferences dialog
        # with three options: never, daily, and weekly. Check that first, but allow it to be
        # overridden by a more specific DaysBetweenUpdates selection, if the user has provided it.
        # For that parameter we use:
        # -1: Only manual updates (default)
        #  0: Update every launch
        # >0: Update every n days
        pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")
        self.update_cache = False
        update_frequency = pref.GetInt("UpdateFrequencyComboEntry", 0)
        if update_frequency == 0:
            days_between_updates = -1
        elif update_frequency == 1:
            days_between_updates = 1
        elif update_frequency == 2:
            days_between_updates = 7
        days_between_updates = pref.GetInt("DaysBetweenUpdates", days_between_updates)
        last_cache_update_string = pref.GetString("LastCacheUpdate", "never")
        cache_path = FreeCAD.getUserCachePath()
        am_path = os.path.join(cache_path, "AddonManager")
        if last_cache_update_string == "never":
            self.update_cache = True
        elif days_between_updates > 0:
            if hasattr(date, "fromisoformat"):
                last_cache_update = date.fromisoformat(last_cache_update_string)
            else:
                # Python 3.6 and earlier don't have date.fromisoformat
                date_re = re.compile("([0-9]{4})-?(1[0-2]|0[1-9])-?(3[01]|0[1-9]|[12][0-9])")
                matches = date_re.match(last_cache_update_string)
                last_cache_update = date(
                    int(matches.group(1)), int(matches.group(2)), int(matches.group(3))
                )
            delta_update = timedelta(days=days_between_updates)
            if date.today() >= last_cache_update + delta_update:
                self.update_cache = True
        elif days_between_updates == 0:
            self.update_cache = True
        elif not os.path.isdir(am_path):
            self.update_cache = True
        stopfile = utils.get_cache_file_name("CACHE_UPDATE_INTERRUPTED")
        if os.path.exists(stopfile):
            self.update_cache = True
            os.remove(stopfile)
            FreeCAD.Console.PrintMessage(
                translate(
                    "AddonsInstaller",
                    "Previous cache process was interrupted, restarting...\n",
                )
            )

        # See if the user has changed the custom repos list since our last re-cache:
        stored_hash = pref.GetString("CustomRepoHash", "")
        custom_repos = pref.GetString("CustomRepositories", "")
        if custom_repos:
            hasher = hashlib.sha1()
            hasher.update(custom_repos.encode("utf-8"))
            new_hash = hasher.hexdigest()
        else:
            new_hash = ""
        if new_hash != stored_hash:
            stored_hash = pref.SetString("CustomRepoHash", new_hash)
            self.update_cache = True
            FreeCAD.Console.PrintMessage(
                translate(
                    "AddonsInstaller",
                    "Custom repo list changed, forcing recache...\n",
                )
            )

    def reject(self) -> None:
        """called when the window has been closed"""

        # save window geometry for next use
        pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")
        pref.SetInt("WindowWidth", self.dialog.width())
        pref.SetInt("WindowHeight", self.dialog.height())

        # ensure all threads are finished before closing
        oktoclose = True
        worker_killed = False
        self.startup_sequence = []
        for worker in self.workers:
            if hasattr(self, worker):
                thread = getattr(self, worker)
                if thread:
                    if not thread.isFinished():
                        thread.blockSignals(True)
                        thread.requestInterruption()
                        worker_killed = True
                        oktoclose = False
        while not oktoclose:
            oktoclose = True
            for worker in self.workers:
                if hasattr(self, worker):
                    thread = getattr(self, worker)
                    if thread:
                        thread.wait(25)
                        if not thread.isFinished():
                            oktoclose = False
            QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents)

        # Write the cache data if it's safe to do so:
        if not worker_killed:
            for repo in self.item_model.repos:
                if repo.repo_type == Addon.Kind.MACRO:
                    self.cache_macro(repo)
                else:
                    self.cache_package(repo)
            self.write_package_cache()
            self.write_macro_cache()
        else:
            self.write_cache_stopfile()
            FreeCAD.Console.PrintLog(
                "Not writing the cache because a process was forcibly terminated and the state is "
                "unknown.\n"
            )

        if self.restart_required:
            # display restart dialog
            m = QtWidgets.QMessageBox()
            m.setWindowTitle(translate("AddonsInstaller", "Addon manager"))
            m.setWindowIcon(QtGui.QIcon(":/icons/AddonManager.svg"))
            m.setText(
                translate(
                    "AddonsInstaller",
                    "You must restart FreeCAD for changes to take effect.",
                )
            )
            m.setIcon(m.Warning)
            m.setStandardButtons(m.Ok | m.Cancel)
            m.setDefaultButton(m.Cancel)
            okBtn = m.button(QtWidgets.QMessageBox.StandardButton.Ok)
            cancelBtn = m.button(QtWidgets.QMessageBox.StandardButton.Cancel)
            okBtn.setText(translate("AddonsInstaller", "Restart now"))
            cancelBtn.setText(translate("AddonsInstaller", "Restart later"))
            ret = m.exec_()
            if ret == m.Ok:
                # restart FreeCAD after a delay to give time to this dialog to close
                QtCore.QTimer.singleShot(1000, utils.restart_freecad)

    def startup(self) -> None:
        """Downloads the available packages listings and populates the table

        This proceeds in four stages: first, the main GitHub repository is queried for a list of
        possible addons. Each addon is specified as a git submodule with name and branch
        information. The actual specific commit ID of the submodule (as listed on Github) is
        ignored. Any extra repositories specified by the user are appended to this list.

        Second, the list of macros is downloaded from the FreeCAD/FreeCAD-macros repository and
        the wiki.

        Third, each of these items is queried for a package.xml metadata file. If that file exists
        it is downloaded, cached, and any icons that it references are also downloaded and cached.

        Finally, for workbenches that are not contained within a package (e.g. they provide no
        metadata), an additional git query is made to see if an update is available. Macros are
        checked for file changes.

        Each of these stages is launched in a separate thread to ensure that the UI remains
        responsive, and the operation can be cancelled.

        Each stage is also subject to caching, so may return immediately, if no cache update has
        been requested."""

        # Each function in this list is expected to launch a thread and connect its completion
        # signal to self.do_next_startup_phase, or to shortcut to calling
        # self.do_next_startup_phase if it is not launching a worker
        self.startup_sequence = [
            self.populate_packages_table,
            self.activate_table_widgets,
            self.populate_macros,
            self.update_metadata_cache,
            self.check_updates,
            self.check_python_updates,
        ]
        pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")
        if pref.GetBool("DownloadMacros", False):
            self.startup_sequence.append(self.load_macro_metadata)
        selection = pref.GetString("SelectedAddon", "")
        if selection:
            self.startup_sequence.insert(2, functools.partial(self.select_addon, selection))
            pref.SetString("SelectedAddon", "")
        self.current_progress_region = 0
        self.number_of_progress_regions = len(self.startup_sequence)
        self.do_next_startup_phase()

    def do_next_startup_phase(self) -> None:
        """Pop the top item in self.startup_sequence off the list and run it"""

        if len(self.startup_sequence) > 0:
            phase_runner = self.startup_sequence.pop(0)
            self.current_progress_region += 1
            phase_runner()
        else:
            self.hide_progress_widgets()
            self.update_cache = False
            self.dialog.buttonUpdateCache.setEnabled(True)
            self.dialog.buttonUpdateCache.setText(
                translate("AddonsInstaller", "Refresh local cache")
            )
            pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")
            pref.SetString("LastCacheUpdate", date.today().isoformat())
            self.packageList.item_filter.invalidateFilter()

    def populate_packages_table(self) -> None:
        self.item_model.clear()

        use_cache = not self.update_cache
        if use_cache:
            if os.path.isfile(utils.get_cache_file_name("package_cache.json")):
                with open(utils.get_cache_file_name("package_cache.json"), encoding="utf-8") as f:
                    data = f.read()
                    try:
                        from_json = json.loads(data)
                        if len(from_json) == 0:
                            use_cache = False
                    except json.JSONDecodeError:
                        use_cache = False
            else:
                use_cache = False

        if not use_cache:
            self.update_cache = True  # Make sure to trigger the other cache updates, if the json
            # file was missing
            self.create_addon_list_worker = CreateAddonListWorker()
            self.create_addon_list_worker.status_message.connect(self.show_information)
            self.create_addon_list_worker.addon_repo.connect(self.add_addon_repo)
            self.update_progress_bar(10, 100)
            self.create_addon_list_worker.finished.connect(
                self.do_next_startup_phase
            )  # Link to step 2
            self.create_addon_list_worker.start()
        else:
            self.create_addon_list_worker = LoadPackagesFromCacheWorker(
                utils.get_cache_file_name("package_cache.json")
            )
            self.create_addon_list_worker.addon_repo.connect(self.add_addon_repo)
            self.update_progress_bar(10, 100)
            self.create_addon_list_worker.finished.connect(
                self.do_next_startup_phase
            )  # Link to step 2
            self.create_addon_list_worker.start()

    def cache_package(self, repo: Addon):
        if not hasattr(self, "package_cache"):
            self.package_cache = {}
        self.package_cache[repo.name] = repo.to_cache()

    def write_package_cache(self):
        if hasattr(self, "package_cache"):
            package_cache_path = utils.get_cache_file_name("package_cache.json")
            with open(package_cache_path, "w", encoding="utf-8") as f:
                f.write(json.dumps(self.package_cache, indent="  "))

    def activate_table_widgets(self) -> None:
        self.packageList.setEnabled(True)
        self.packageList.ui.lineEditFilter.setFocus()
        self.do_next_startup_phase()

    def populate_macros(self) -> None:
        macro_cache_file = utils.get_cache_file_name("macro_cache.json")
        cache_is_bad = True
        if os.path.isfile(macro_cache_file):
            size = os.path.getsize(macro_cache_file)
            if size > 1000:  # Make sure there is actually data in there
                cache_is_bad = False
        if cache_is_bad:
            if not self.update_cache:
                self.update_cache = True  # Make sure to trigger the other cache updates, if the
                # json file was missing
                self.create_addon_list_worker = CreateAddonListWorker()
                self.create_addon_list_worker.status_message.connect(self.show_information)
                self.create_addon_list_worker.addon_repo.connect(self.add_addon_repo)
                self.update_progress_bar(10, 100)
                self.create_addon_list_worker.finished.connect(
                    self.do_next_startup_phase
                )  # Link to step 2
                self.create_addon_list_worker.start()
            else:
                # It's already been done in the previous step (TODO: Refactor to eliminate this
                # step)
                self.do_next_startup_phase()
        else:
            self.macro_worker = LoadMacrosFromCacheWorker(
                utils.get_cache_file_name("macro_cache.json")
            )
            self.macro_worker.add_macro_signal.connect(self.add_addon_repo)
            self.macro_worker.finished.connect(self.do_next_startup_phase)
            self.macro_worker.start()

    def cache_macro(self, repo: Addon):
        if not hasattr(self, "macro_cache"):
            self.macro_cache = []
        if repo.macro is not None:
            self.macro_cache.append(repo.macro.to_cache())
        else:
            FreeCAD.Console.PrintError(
                f"Addon Manager: Internal error, cache_macro called on non-macro {repo.name}\n"
            )

    def write_macro_cache(self):
        if not hasattr(self, "macro_cache"):
            return
        macro_cache_path = utils.get_cache_file_name("macro_cache.json")
        with open(macro_cache_path, "w", encoding="utf-8") as f:
            f.write(json.dumps(self.macro_cache, indent="  "))
            self.macro_cache = []

    def update_metadata_cache(self) -> None:
        if self.update_cache:
            self.update_metadata_cache_worker = UpdateMetadataCacheWorker(self.item_model.repos)
            self.update_metadata_cache_worker.status_message.connect(self.show_information)
            self.update_metadata_cache_worker.finished.connect(
                self.do_next_startup_phase
            )  # Link to step 4
            self.update_metadata_cache_worker.progress_made.connect(self.update_progress_bar)
            self.update_metadata_cache_worker.package_updated.connect(self.on_package_updated)
            self.update_metadata_cache_worker.start()
        else:
            self.do_next_startup_phase()

    def on_buttonUpdateCache_clicked(self) -> None:
        self.update_cache = True
        cache_path = FreeCAD.getUserCachePath()
        am_path = os.path.join(cache_path, "AddonManager")
        utils.rmdir(am_path)
        self.dialog.buttonUpdateCache.setEnabled(False)
        self.dialog.buttonUpdateCache.setText(translate("AddonsInstaller", "Updating cache..."))
        self.startup()

        # Recaching implies checking for updates, regardless of the user's autocheck option
        self.startup_sequence.remove(self.check_updates)
        self.startup_sequence.append(self.force_check_updates)

    def on_package_updated(self, repo: Addon) -> None:
        """Called when the named package has either new metadata or a new icon (or both)"""

        with self.lock:
            repo.icon = self.get_icon(repo, update=True)
            self.item_model.reload_item(repo)

    def load_macro_metadata(self) -> None:
        if self.update_cache:
            self.load_macro_metadata_worker = CacheMacroCodeWorker(self.item_model.repos)
            self.load_macro_metadata_worker.status_message.connect(self.show_information)
            self.load_macro_metadata_worker.update_macro.connect(self.on_package_updated)
            self.load_macro_metadata_worker.progress_made.connect(self.update_progress_bar)
            self.load_macro_metadata_worker.finished.connect(self.do_next_startup_phase)
            self.load_macro_metadata_worker.start()
        else:
            self.do_next_startup_phase()

    def select_addon(self, name: str) -> None:
        found = False
        for addon in self.item_model.repos:
            if addon.name == name:
                self.table_row_activated(addon)
                found = True
                break
        if not found:
            FreeCAD.Console.PrintWarning(
                translate("AddonsInstaller", "Could not find addon '{}' to select\n").format(name)
            )
        self.do_next_startup_phase()

    def check_updates(self) -> None:
        "checks every installed addon for available updates"

        pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")
        autocheck = pref.GetBool("AutoCheck", False)
        if not autocheck:
            FreeCAD.Console.PrintLog(
                "Addon Manager: Skipping update check because AutoCheck user preference is False\n"
            )
            self.do_next_startup_phase()
            return
        if not self.packages_with_updates:
            self.force_check_updates(standalone=False)
        else:
            self.do_next_startup_phase()

    def force_check_updates(self, standalone=False) -> None:
        if hasattr(self, "check_worker"):
            thread = self.check_worker
            if thread:
                if not thread.isFinished():
                    self.do_next_startup_phase()
                    return

        self.dialog.buttonUpdateAll.setText(translate("AddonsInstaller", "Checking for updates..."))
        self.packages_with_updates.clear()
        self.dialog.buttonUpdateAll.show()
        self.dialog.buttonCheckForUpdates.setDisabled(True)
        self.check_worker = CheckWorkbenchesForUpdatesWorker(self.item_model.repos)
        self.check_worker.finished.connect(self.do_next_startup_phase)
        self.check_worker.finished.connect(self.update_check_complete)
        self.check_worker.progress_made.connect(self.update_progress_bar)
        if standalone:
            self.current_progress_region = 1
            self.number_of_progress_regions = 1
        self.check_worker.update_status.connect(self.status_updated)
        self.check_worker.start()
        self.enable_updates(len(self.packages_with_updates))

    def status_updated(self, repo: Addon) -> None:
        self.item_model.reload_item(repo)
        if repo.status() == Addon.Status.UPDATE_AVAILABLE:
            self.packages_with_updates.add(repo)
            self.enable_updates(len(self.packages_with_updates))
        elif repo.status() == Addon.Status.PENDING_RESTART:
            self.restart_required = True

    def enable_updates(self, number_of_updates: int) -> None:
        """enables the update button"""

        if number_of_updates:
            s = translate("AddonsInstaller", "Apply {} update(s)", "", number_of_updates)
            self.dialog.buttonUpdateAll.setText(s.format(number_of_updates))
            self.dialog.buttonUpdateAll.setEnabled(True)
        elif hasattr(self, "check_worker") and self.check_worker.isRunning():
            self.dialog.buttonUpdateAll.setText(
                translate("AddonsInstaller", "Checking for updates...")
            )
        else:
            self.dialog.buttonUpdateAll.setText(
                translate("AddonsInstaller", "No updates available")
            )
            self.dialog.buttonUpdateAll.setEnabled(False)

    def update_check_complete(self) -> None:
        self.enable_updates(len(self.packages_with_updates))
        self.dialog.buttonCheckForUpdates.setEnabled(True)

    def check_python_updates(self) -> None:
        PythonPackageManager.migrate_old_am_installations()  # Migrate 0.20 to 0.21
        self.do_next_startup_phase()

    def show_python_updates_dialog(self) -> None:
        if not hasattr(self, "manage_python_packages_dialog"):
            self.manage_python_packages_dialog = PythonPackageManager(self.item_model.repos)
        self.manage_python_packages_dialog.show()

    def show_developer_tools(self) -> None:
        """Display the developer tools dialog"""
        if not self.developer_mode:
            self.developer_mode = DeveloperMode()
        self.developer_mode.show()

        checker = MetadataValidators()
        checker.validate_all(self.item_model.repos)

    def add_addon_repo(self, addon_repo: Addon) -> None:
        """adds a workbench to the list"""

        if addon_repo.icon is None or addon_repo.icon.isNull():
            addon_repo.icon = self.get_icon(addon_repo)
        for repo in self.item_model.repos:
            if repo.name == addon_repo.name:
                # self.item_model.reload_item(repo) # If we want to have later additions superseded
                # earlier
                return
        self.item_model.append_item(addon_repo)

    def get_icon(self, repo: Addon, update: bool = False) -> QtGui.QIcon:
        """Returns an icon for an Addon. Uses a cached icon if possible, unless update is True,
        in which case the icon is regenerated."""

        if not update and repo.icon and not repo.icon.isNull() and repo.icon.isValid():
            return repo.icon

        path = ":/icons/" + repo.name.replace(" ", "_")
        if repo.repo_type == Addon.Kind.WORKBENCH:
            path += "_workbench_icon.svg"
            default_icon = QtGui.QIcon(":/icons/document-package.svg")
        elif repo.repo_type == Addon.Kind.MACRO:
            if repo.macro and repo.macro.icon:
                if os.path.isabs(repo.macro.icon):
                    path = repo.macro.icon
                    default_icon = QtGui.QIcon(":/icons/document-python.svg")
                else:
                    path = os.path.join(os.path.dirname(repo.macro.src_filename), repo.macro.icon)
                    default_icon = QtGui.QIcon(":/icons/document-python.svg")
            elif repo.macro and repo.macro.xpm:
                cache_path = FreeCAD.getUserCachePath()
                am_path = os.path.join(cache_path, "AddonManager", "MacroIcons")
                os.makedirs(am_path, exist_ok=True)
                path = os.path.join(am_path, repo.name + "_icon.xpm")
                if not os.path.exists(path):
                    with open(path, "w") as f:
                        f.write(repo.macro.xpm)
                default_icon = QtGui.QIcon(repo.macro.xpm)
            else:
                path += "_macro_icon.svg"
                default_icon = QtGui.QIcon(":/icons/document-python.svg")
        elif repo.repo_type == Addon.Kind.PACKAGE:
            # The cache might not have been downloaded yet, check to see if it's there...
            if os.path.isfile(repo.get_cached_icon_filename()):
                path = repo.get_cached_icon_filename()
            elif repo.contains_workbench():
                path += "_workbench_icon.svg"
                default_icon = QtGui.QIcon(":/icons/document-package.svg")
            elif repo.contains_macro():
                path += "_macro_icon.svg"
                default_icon = QtGui.QIcon(":/icons/document-python.svg")
            else:
                default_icon = QtGui.QIcon(":/icons/document-package.svg")

        if QtCore.QFile.exists(path):
            addonicon = QtGui.QIcon(path)
        else:
            addonicon = default_icon
        repo.icon = addonicon

        return addonicon

    def table_row_activated(self, selected_repo: Addon) -> None:
        """a row was activated, show the relevant data"""

        self.packageList.hide()
        self.packageDetails.show()
        self.packageDetails.show_repo(selected_repo)

    def show_information(self, message: str) -> None:
        """shows generic text in the information pane"""

        self.dialog.labelStatusInfo.setText(message)
        self.dialog.labelStatusInfo.repaint()

    def show_workbench(self, repo: Addon) -> None:
        self.packageList.hide()
        self.packageDetails.show()
        self.packageDetails.show_repo(repo)

    def on_buttonBack_clicked(self) -> None:
        self.packageDetails.hide()
        self.packageList.show()

    def append_to_repos_list(self, repo: Addon) -> None:
        """this function allows threads to update the main list of workbenches"""
        self.item_model.append_item(repo)

    def update(self, repo: Addon) -> None:
        self.launch_installer_gui(repo)

    def mark_repo_update_available(self, repo: Addon, available: bool) -> None:
        if available:
            repo.set_status(Addon.Status.UPDATE_AVAILABLE)
        else:
            repo.set_status(Addon.Status.NO_UPDATE_AVAILABLE)
        self.item_model.reload_item(repo)
        self.packageDetails.show_repo(repo)

    def launch_installer_gui(self, addon: Addon) -> None:
        if self.installer_gui is not None:
            FreeCAD.Console.PrintError(
                translate(
                    "AddonsInstaller",
                    "Cannot launch a new installer until the previous one has finished.",
                )
            )
            return
        if addon.macro is not None:
            self.installer_gui = MacroInstallerGUI(addon)
        else:
            self.installer_gui = AddonInstallerGUI(addon, self.item_model.repos)
        self.installer_gui.success.connect(self.on_package_status_changed)
        self.installer_gui.finished.connect(self.cleanup_installer)
        self.installer_gui.run()  # Does not block

    def cleanup_installer(self) -> None:
        QtCore.QTimer.singleShot(500, self.no_really_clean_up_the_installer)

    def no_really_clean_up_the_installer(self) -> None:
        self.installer_gui = None

    def update_all(self) -> None:
        """Asynchronously apply all available updates: individual failures are noted, but do not
        stop other updates"""

        if self.installer_gui is not None:
            FreeCAD.Console.PrintError(
                translate(
                    "AddonsInstaller",
                    "Cannot launch a new installer until the previous one has finished.",
                )
            )
            return

        self.installer_gui = UpdateAllGUI(self.item_model.repos)
        self.installer_gui.addon_updated.connect(self.on_package_status_changed)
        self.installer_gui.finished.connect(self.cleanup_installer)
        self.installer_gui.run()  # Does not block

    def hide_progress_widgets(self) -> None:
        """hides the progress bar and related widgets"""

        self.dialog.labelStatusInfo.hide()
        self.dialog.progressBar.hide()
        self.dialog.buttonPauseUpdate.hide()
        self.packageList.ui.lineEditFilter.setFocus()

    def show_progress_widgets(self) -> None:
        if self.dialog.progressBar.isHidden():
            self.dialog.progressBar.show()
            self.dialog.buttonPauseUpdate.show()
            self.dialog.labelStatusInfo.show()

    def update_progress_bar(self, current_value: int, max_value: int) -> None:
        """Update the progress bar, showing it if it's hidden"""

        max_value = max_value if max_value > 0 else 1

        if current_value < 0:
            current_value = 0
        elif current_value > max_value:
            current_value = max_value

        self.show_progress_widgets()
        region_size = 100.0 / self.number_of_progress_regions
        completed_region_portion = (self.current_progress_region - 1) * region_size
        current_region_portion = (float(current_value) / float(max_value)) * region_size
        value = completed_region_portion + current_region_portion
        self.dialog.progressBar.setValue(
            value * 10
        )  # Out of 1000 segments, so it moves sort of smoothly
        self.dialog.progressBar.repaint()

    def stop_update(self) -> None:
        self.cleanup_workers()
        self.hide_progress_widgets()
        self.write_cache_stopfile()
        self.dialog.buttonUpdateCache.setEnabled(True)
        self.dialog.buttonUpdateCache.setText(translate("AddonsInstaller", "Refresh local cache"))

    def write_cache_stopfile(self) -> None:
        stopfile = utils.get_cache_file_name("CACHE_UPDATE_INTERRUPTED")
        with open(stopfile, "w", encoding="utf8") as f:
            f.write(
                "This file indicates that a cache operation was interrupted, and "
                "the cache is in an unknown state. It will be deleted next time "
                "AddonManager recaches."
            )

    def on_package_status_changed(self, repo: Addon) -> None:
        if repo.status() == Addon.Status.PENDING_RESTART:
            self.restart_required = True
        self.item_model.reload_item(repo)
        self.packageDetails.show_repo(repo)
        if repo in self.packages_with_updates:
            self.packages_with_updates.remove(repo)
            self.enable_updates(len(self.packages_with_updates))

    def executemacro(self, repo: Addon) -> None:
        """executes a selected macro"""

        macro = repo.macro
        if not macro or not macro.code:
            return

        if macro.is_installed():
            macro_path = os.path.join(self.macro_repo_dir, macro.filename)
            FreeCADGui.open(str(macro_path))
            self.dialog.hide()
            FreeCADGui.SendMsgToActiveView("Run")
        else:
            with tempfile.TemporaryDirectory() as dir:
                temp_install_succeeded = macro.install(dir)
                if not temp_install_succeeded:
                    message = translate(
                        "AddonsInstaller",
                        "Execution of macro failed. See console for failure details.",
                    )
                    return
                macro_path = os.path.join(dir, macro.filename)
                FreeCADGui.open(str(macro_path))
                self.dialog.hide()
                FreeCADGui.SendMsgToActiveView("Run")

    def remove(self, addon: Addon) -> None:
        """Remove this addon."""
        if self.installer_gui is not None:
            FreeCAD.Console.PrintError(
                translate(
                    "AddonsInstaller",
                    "Cannot launch a new installer until the previous one has finished.",
                )
            )
            return
        self.installer_gui = AddonUninstallerGUI(addon)
        self.installer_gui.finished.connect(self.cleanup_installer)
        self.installer_gui.finished.connect(
            functools.partial(self.on_package_status_changed, addon)
        )
        self.installer_gui.run()  # Does not block


# @}
